from typing import Callable

import torch

from goldener.vectorize import TensorVectorizer, Filter2DWithCount, FilterLocation


def get_vit_class_token_vectorizer() -> TensorVectorizer:
    """Get a TensorVectorizer that keeps only the class token from a ViT model."""
    return TensorVectorizer(
        keep=Filter2DWithCount(
            filter_count=1, keep=True, filter_location=FilterLocation.START
        ),
        channel_pos=2,
    )


def get_vit_prefix_tokens_vectorizer(n_prefixes: int = 5) -> TensorVectorizer:
    """Get a TensorVectorizer that keeps the first n_prefixes tokens from a ViT model.

    Args:
        n_prefixes: The number of prefix tokens to keep (default is 5).
    """
    if n_prefixes <= 0:
        raise ValueError("n_prefixes must be a positive integer")

    return TensorVectorizer(
        keep=Filter2DWithCount(
            filter_count=n_prefixes, keep=True, filter_location=FilterLocation.START
        ),
        channel_pos=2,
    )


def get_vit_patch_tokens_vectorizer(
    n_prefixes: int | None = 5,
    n_random: int | None = None,
    transform_y: Callable[[torch.Tensor], torch.Tensor] | None = None,
    generator: torch.Generator | None = None,
) -> TensorVectorizer:
    """Get a TensorVectorizer that keeps the patch tokens from a ViT model

    It optionally removes the first n_prefixes tokens and keeps n_random random tokens.

    Args:
        n_prefixes: The number of prefix tokens to remove from the start (default is 5).
           If None, no prefixes are removed.
        n_random: The number of random tokens to keep after removing prefixes
           (default is None, meaning no random tokens are kept).
           If None, no random filtering is applied.
        transform_y: An optional callable to transform the target variable y.
        generator: An optional torch.Generator for reproducibility when n_random is set.
    """
    if n_prefixes is not None and n_prefixes <= 0:
        raise ValueError("n_prefixes must be a positive integer or None")
    if n_random is not None and n_random <= 0:
        raise ValueError("n_random must be a positive integer or None")

    return TensorVectorizer(
        remove=(
            Filter2DWithCount(
                filter_count=n_prefixes,
                keep=False,
                filter_location=FilterLocation.START,
            )
            if n_prefixes is not None
            else None
        ),
        random=(
            Filter2DWithCount(
                filter_count=n_random,
                keep=True,
                filter_location=FilterLocation.RANDOM,
                generator=generator,
            )
            if n_random is not None
            else None
        ),
        channel_pos=2,
        transform_y=transform_y,
    )



import math
from typing import Iterable, Any, TypeVar

import torch

from goldener.torch_utils import get_unique_values_in_tensor


def check_x_and_y_shapes(x_shape: tuple[int, ...], y_shape: tuple[int, ...]) -> None:
    """Check compatibility of shapes of x and y tensors.

    Args:
        x_shape: Shape of the input tensor x.
        y_shape: Shape of the target tensor y.

    Raises:
        ValueError: If the shapes are incompatible. See the conditions below for details.
    Conditions:
        - If x is 1D, x and y must have the same shape.
        - If x is 2D or more, y must have a channel dimension of 1 (i.e., y_shape[1] == 1).
        - If y has a batch size greater than 1 (i.e., y_shape[0] > 1), x and y must have the same batch size (i.e., x_shape[0] == y_shape[0]).
        - If x has more than 2 dimensions, the additional (after channel) dimensions of x and y must match.
    """

    if len(x_shape) == 1:
        if x_shape != y_shape:
            raise ValueError("x and y must have the same shape when x is 1D")
    else:
        if y_shape[1] != 1:
            raise ValueError(
                "x and y must have the same channel dimension when x is 2D"
            )

        batch_y = y_shape[0]
        if batch_y > 1 and x_shape[0] != batch_y:
            raise ValueError(
                "x and y must have the same batch size when y batch size > 1"
            )

        if len(x_shape) > 2 and x_shape[2:] != y_shape[2:]:
            raise ValueError(
                "x and y must have compatible shapes when x is more than 2D"
            )


def get_ratio_list_sum(ratios: list[float]) -> float:
    """Get the sum of a list of ratios and validate it (total between 0 and 1).

    Args:
        ratios: A list of float ratios.

    Returns:
        The sum of the ratios.

    Raises:
        ValueError: If the sum of the ratios is not greater than 0 and at most 1.
    """
    ratio_sum = sum(ratios)
    if not all(ratio >= 0 for ratio in ratios):
        raise ValueError("Ratios must be non-negative.")

    if not (0 < ratio_sum <= 1.0):
        raise ValueError("Sum of ratios must be 1.0.")

    return ratio_sum


def get_ratios_for_counts(counts: list[int]) -> list[float]:
    """Get ratios for a list of counts.

    Args:
        counts: A list of integer counts.

    Returns:
        A list of float ratios corresponding to the input counts.
    """
    if len(counts) == 0:
        raise ValueError("Counts list cannot be empty.")

    if len(counts) == 1:
        return [1.0]  # single count gets all the ratio

    total = sum(counts)
    if total == 0:
        raise ValueError("Total count must be greater than 0.")

    first_ratios = [count / total for count in counts[:-1]]

    return first_ratios + [
        1.0 - sum(first_ratios)
    ]  # ensure the last ratio is adjusted to sum to 1.0


def split_sampling_among_chunks(to_split: int, chunk_sizes: list[int]) -> list[int]:
    """Split a total sampling count among chunks based on their sizes.

    Args:
        to_split: The total number of samples to split among the chunks.
        chunk_sizes: A list of integers representing the sizes of each chunk.

    Returns:
        A list of integers representing the number of samples to draw from each chunk, summing up to to_split.

    Raises:
        ValueError: If the input parameters are invalid. See the conditions below for details.
    Conditions:
        - At least one chunk size must be provided.
        - If only one chunk size is provided, to_split must not be greater than that chunk size.
        - The total size of the chunks must be greater than 0.
    """
    if len(chunk_sizes) == 0:
        raise ValueError("At least one chunk size is required to split the sampling")

    if len(chunk_sizes) == 1:
        if to_split > chunk_sizes[0]:
            raise ValueError(
                f"Split count {to_split} cannot be greater than chunk size {chunk_sizes[0]}"
            )
        return [to_split]

    total_size = sum(chunk_sizes)
    if total_size == 0:
        raise ValueError("Total size of chunks must be greater than 0.")

    ratios = get_ratios_for_counts(chunk_sizes)
    counts = get_sampling_count_from_ratios(
        ratios={
            i: ratio for i, ratio in enumerate(ratios)
        },  # add dummy keys for the ratios
        sampling_size=to_split,
        force_non_zero=False,
    )
    counts = dict(sorted(counts.items()))  # reorder in the chunk order

    return list(counts.values())


def filter_batch_from_indices(
    batch: dict[str, Any],
    to_remove: set[int],
    index_key: str = "idx",
) -> dict[str, Any]:
    """Filter a batch dictionary to only include items at specified indices.

    Args:
        batch: A dictionary representing a batch of data (each key corresponds to stacked information).
        to_remove: A set of indices to filter out from the batch.
        index_key: The key in the batch dictionary that contains the indices.

    Returns:
        A filtered batch dictionary without the specified indices.
    """
    keep_in_batch = [
        idx_position
        for idx_position, idx_value in enumerate(batch[index_key])
        if (idx_value.item() if isinstance(idx_value, torch.Tensor) else idx_value)
        not in to_remove
    ]
    if not keep_in_batch:
        return {}  # all samples already described

    def filter_batched_values(
        batched_value: list | torch.Tensor,
    ) -> list | torch.Tensor:
        """Inner function to remove already described samples from the batch."""
        filtered = [
            value
            for idx_value, value in enumerate(batched_value)
            if idx_value in keep_in_batch
        ]
        if isinstance(batched_value, torch.Tensor):
            return torch.stack(filtered, dim=0)
        else:
            return filtered

    return {
        key: filter_batched_values(batched_value)
        for key, batched_value in batch.items()
    }


def get_indices_with_labels(
    batch: dict[str, Any],
    label_key: str,
    exclude_labels: set[str],
    index_key: str = "idx",
) -> set[int]:
    """Get the set of index values for batch items whose sharing label values with exclude_labels.

    Args:
        batch: A dictionary representing a batch of data. It must contain the keys specified by label_key and index_key,
        where the values are lists or tensors of the same length. Label values can be strings or iterables of strings.
        label_key: The key in the batch dictionary that contains the labels.
        exclude_labels: A set of label strings to exclude.
        index_key: The key in the batch dictionary that contains the indices.

    Returns:
        A set of index values corresponding to items with excluded labels.
    """
    indices = set()
    for idx_value, label_value in zip(batch[index_key], batch[label_key]):
        labels = (
            {
                label_value,
            }
            if isinstance(label_value, str)
            else label_value
        )
        if exclude_labels.intersection(labels):
            indices.add(
                int(idx_value.item())
                if isinstance(idx_value, torch.Tensor)
                else idx_value
            )

    return indices


def get_size_and_sampling_count_per_chunk(
    total_size: int, sampling_size: int, max_chunk_size: int
) -> tuple[list[int], list[int]]:
    """Get sizes and sampling sizes per chunk.

    Args:
        total_size: Total size of the data to sample from.
        sampling_size: Total number of samples to draw.
        max_chunk_size: Maximum size of each chunk.

    Returns:
        A tuple of two lists:
            - A list of integers representing the size of each chunk.
            - A list of integers representing the number of samples to draw from each chunk.

    Raises:
        ValueError: If sampling_size is greater than or equal to total_size.
    """
    if sampling_size >= total_size:
        raise ValueError("sampling_size must be less than or equal to total_size")

    if max_chunk_size >= total_size:
        return [total_size], [sampling_size]  # single chunk

    chunk_count = math.ceil(total_size / max_chunk_size)
    chunk_size = total_size // chunk_count
    chunk_sampling_size = sampling_size // chunk_count

    chunk_sizes = [chunk_size] * (chunk_count - 1)
    chunk_sizes.append(total_size - sum(chunk_sizes))
    chunk_sampling_sizes = [chunk_sampling_size] * (chunk_count - 1)
    chunk_sampling_sizes.append(sampling_size - sum(chunk_sampling_sizes))

    return chunk_sizes, chunk_sampling_sizes


def check_sampling_size(
    sampling_size: int | float, total_size: int | None = None, force_max: bool = False
) -> None:
    """Check the validity of the sampling size.

    Args:
        sampling_size: The sampling size to check (can be int or float).
        total_size: The total size of the data (Optional).
        force_max: Whether the sampling size is expected to be the maximum possible (1.0 for float, total for int)

    Raises:
        ValueError: If the sampling size is invalid based on its type and total size.
        If sampling_size is a float, it must be in the range (0.0, 1.0].
        If sampling_size is an int, it must be in the range (0, total_size].
    """
    if isinstance(sampling_size, float):
        if force_max:
            if not math.isclose(sampling_size, 1.0, rel_tol=1e-9, abs_tol=0.0):
                raise ValueError("Sampling size as float must be equal to 1.0")
        else:
            if not 0 < sampling_size <= 1.0:
                raise ValueError(
                    "Sampling size as float must be greater than 0.0 and at most 1.0"
                )

    if isinstance(sampling_size, int) and total_size is not None:
        if force_max:
            if sampling_size != total_size:
                raise ValueError(
                    "Sampling size as int must be equal to the total number of samples"
                )
        else:
            if not 0 < sampling_size <= total_size:
                raise ValueError(
                    "Sampling size as int must be greater than 0 and less or equal than the total number of samples"
                )


def check_all_same_type(iterable: Iterable[Any]) -> None:
    """Check if all elements in an iterable are of the same type.

    Args:
        iterable: An iterable containing elements to check.

    Raises:
        TypeError: If any element in the iterable is not of the expected type.
    """
    first_type = type(next(iter(iterable)))
    different = []
    for item in iterable:
        if not isinstance(item, first_type):
            different.append(type(item))

    if different:
        raise TypeError(
            f"All elements must be of the same type {first_type}. Found different types: {different}"
        )


def get_sampling_count_from_size(
    sampling_size: int | float, total_size: int | None = None
) -> int:
    """Get the sampling count from the sampling size.

    Args:
        sampling_size: The sampling size (can be int or float). If int, it represents the exact number of samples
        and it is required to be less than total_size. If float, it represents the fraction of total size and it
        is required to be in the range (0.0, 1.0].
        total_size: The total size of the data (Optional, required if sampling_size is float).

    Returns:
        The calculated sampling count as an integer.

    Raises:
        ValueError: If the sampling size is invalid based on its type and total size.
    """
    if isinstance(sampling_size, int):
        if sampling_size <= 0:
            raise ValueError("Sampling size as int must be greater than 0.")

        if total_size is not None and not (sampling_size < total_size):
            raise ValueError(
                "Sampling size as int must be less than the total number of samples."
            )

        return sampling_size

    if not (0 < sampling_size <= 1.0):
        raise ValueError(
            "Sampling size as float must be greater than 0.0 and at most 1.0."
        )

    if total_size is None:
        raise ValueError("Total size must be provided when sampling size is a float.")

    return math.ceil(sampling_size * total_size)


def transform_batch_from_multiple_to_binarized_targets(
    batch: dict[str, Any],
    target_key: str,
    label_key: str | None,
    target_to_label: dict[tuple[int, ...], str],
    exclude_labels: set[str] | None = None,
    exclude_full_zero: bool = False,
) -> dict[str, Any]:
    """Transform a batch with multiple targets into a batch with binarized targets for each label.

    Args:
        batch: A dictionary representing a batch of data, where the target key contains a target with multiple
            values corresponding to the labels in target_to_label.
        target_key: The key in the batch dictionary that contains the target.
        label_key: The key in the batch dictionary that containes the labels.
        target_to_label: A mapping from unique target tuples to label values.
        exclude_labels: An optional set of labels to exclude from the transformation (default: None).
        exclude_full_zero: Whether to exclude the all-zero target from the transformation (default: False).

    Returns:
        A new batch dictionary where the target key contains binarized targets for each label
        and the label key contains the corresponding labels,
        with other batch elements duplicated accordingly.
    """
    multi_target = batch[target_key]

    multi_target_shape = multi_target.shape
    bin_target_shape = list(multi_target_shape)
    bin_target_shape[1] = 1

    # create a new binarized target from the target
    # containing a different value for each label
    target_flattened = multi_target.movedim(1, -1).reshape(-1, multi_target_shape[1])
    target_per_label = {}
    for unique_target in get_unique_values_in_tensor(target_flattened, -1):
        if exclude_full_zero and (unique_target == 0).all():
            continue  # skip the all-zero target if specified

        unique_target_tuple = tuple(unique_target.tolist())
        if unique_target_tuple not in target_to_label:
            raise ValueError(
                f"Unique target {unique_target} not found in target_to_label mapping."
            )
        label = target_to_label[unique_target_tuple]

        if exclude_labels is not None and label in exclude_labels:
            continue

        new_target = (
            (target_flattened == unique_target.unsqueeze(0))
            .all(dim=1)
            .to(torch.uint8)
            .reshape(bin_target_shape)
        )

        target_per_label[label] = new_target

    if not target_per_label:
        raise ValueError(
            "No valid targets found after applying exclude_full_zero filter."
        )

    # duplicate the batch element for each label/target and
    # insert the corresponding binarized target/label in the batch alongside them.
    new_batch: dict[str, torch.Tensor | list[Any]] = {}
    n_values = len(target_per_label)

    for batch_key, batch_value in batch.items():
        if batch_key not in (target_key, label_key):
            if isinstance(batch_value, torch.Tensor):
                new_batch[batch_key] = torch.cat([batch_value] * n_values, dim=0)
            else:
                if not isinstance(batch_value, list):
                    raise ValueError(
                        f"Batch value for key {batch_key} must be a list or a torch.Tensor. Got {type(batch_value)}."
                    )
                new_batch[batch_key] = batch_value * n_values
        else:
            if batch_key == target_key:
                new_batch[target_key] = torch.cat(
                    [target for target in target_per_label.values()],
                    dim=0,
                )
            elif label_key is not None and batch_key == label_key:
                new_labels = []
                for label, target in target_per_label.items():
                    new_labels += [label] * len(target)
                new_batch[label_key] = new_labels

    # add label if not already there
    if label_key is not None and label_key not in new_batch:
        new_batch[label_key] = list(target_per_label.keys())

    return new_batch


T = TypeVar("T")


def get_sampling_count_from_ratios(
    ratios: dict[T, float], sampling_size: int, force_non_zero: bool = False
) -> dict[T, int]:
    """Get the sampling count for each key from the ratios.

    Args:
        ratios: A dictionary mapping keys to their corresponding ratios (values between 0 and 1).
            the initial counts are computed by multiplying the sampling size by the ratios
            and taking the floor of the result. If the sum of the initial counts is less than
            the sampling size, the remaining samples are distributed one by one starting from the
            most represented key until the total count matches the sampling size.
        sampling_size: The total number of samples to draw.
        force_non_zero: Whether to force a non-zero sampling count for each key (default: False).
            If activated, the least represented keys will have at least 1 sample,
            and the most represented keys will have their count reduced accordingly to maintain the total sampling size.
    Returns:
        A dictionary mapping keys to their corresponding sampling counts (integers).
            The dictionary is sorted by count in ascending order.
    """
    if not math.isclose(sum(ratios.values()), 1.0, rel_tol=1e-9, abs_tol=0.0):
        raise ValueError("Ratios must sum to 1.0")

    # compute the count per key and make sure they sum to the sampling size by
    # adjusting each count starting from the most represented key until the total count matches the sampling size
    counts = {key: math.floor(sampling_size * ratio) for key, ratio in ratios.items()}

    count_total = sum(counts.values())
    if count_total < sampling_size:
        counts = dict(sorted(counts.items(), key=lambda x: x[1], reverse=True))
        to_distribute = sampling_size - count_total
        loop_count = math.ceil(to_distribute / len(counts))
        for _ in range(loop_count):
            for key in counts.keys():
                counts[key] += 1
                to_distribute -= 1
                if to_distribute == 0:
                    break

    assert sum(counts.values()) == sampling_size

    if not force_non_zero:
        return dict(sorted(counts.items(), key=lambda x: x[1]))

    # If force_non_zero is activated, adjust counts to ensure no key has a count of zero
    # first, the zero counts are identified and incremented by 1, while keeping track of how many were incremented
    added = 0
    for key, count in counts.items():
        if count == 0:
            added += 1
            counts[key] += 1

    # then, for each added count, the counts are sorted in descending order
    # and the highest count is decremented by 1 to maintain the total sampling size
    for _ in range(added):
        counts = dict(sorted(counts.items(), key=lambda x: x[1], reverse=True))
        highest_count_key = next(iter(counts.keys()))
        counts[highest_count_key] -= 1
        if counts[highest_count_key] == 0:
            raise ValueError(
                f"While trying to adjust counts, the count for key '{highest_count_key}' became zero."
            )

    return dict(sorted(counts.items(), key=lambda x: x[1]))



import logging
from abc import ABC, abstractmethod

import torch
from torch.nn import Module

from umap import UMAP
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from sklearn.random_projection import GaussianRandomProjection


logger = logging.getLogger(__name__)


class GoldReductionTool(ABC):
    """Reduce the dimensionality of 2D vectors.

    Base class for all dimensionality reduction tools.

    """

    def _validate_input(self, x: torch.Tensor) -> None:
        """Validate that input is already vectorized (2D torch.Tensor).

        Args:
            x: Input tensor to validate.

        Raises:
            ValueError: If input is not a 2D tensor.
        """
        if x.ndim != 2:
            raise ValueError(
                f"GoldReductionTool only accepts 2D tensors (num_vectors, feature_dim). "
                f"Got shape {x.shape}. Please ensure your input is already vectorized."
            )

    @abstractmethod
    def transform(self, x: torch.Tensor) -> torch.Tensor:
        """Transform the 2D vectors using the dimensionality reduction tool.

        Args:
            x: Already vectorized 2D input tensor of shape (num_vectors, feature_dim) to transform.

        Returns:
            Transformed tensor with reduced dimensionality.
        """


class GoldReductionToolWithFit(GoldReductionTool):
    """Reduce the dimensionality of 2D vectors from fittable reduction methods.

    Base class for dimensionality reduction tools that require fitting.
    """

    @abstractmethod
    def fit(self, x: torch.Tensor) -> None:
        """Fit the dimensionality reduction tool to the data.

        Args:
            x: Already vectorized 2D input tensor of shape (num_vectors, feature_dim) to fit the model on.
        """

    @abstractmethod
    def fit_transform(self, x: torch.Tensor) -> torch.Tensor:
        """Fit the dimensionality reduction tool to the 2D vectors and transform it.

        Args:
            x: Already vectorized 2D input tensor of shape (num_vectors, feature_dim) to fit and transform.

        Returns:
            Transformed tensor with reduced dimensionality.
        """


class GoldTorchModuleReductionTool(GoldReductionTool):
    """Dimensionality reduction of 2D vectors with a torch Module.

    Attributes:
        reducer: The torch Module to use for dimensionality reduction.
    """

    def __init__(self, reducer: Module):
        """Initialize the GoldTorchModuleReductionTool.

        Args:
            reducer: The torch Module to use for dimensionality reduction.
        """
        self.reducer = reducer

    def transform(self, x: torch.Tensor) -> torch.Tensor:
        """Transform the 2D vectors using the fitted dimensionality reduction model.

        Args:
            x: x: Already vectorized 2D input tensor of shape (num_vectors, feature_dim) to transform.

        Returns:
            Transformed tensor with reduced dimensionality.
        """
        self._validate_input(x)
        return self.reducer(x)


class GoldSKLearnReductionTool(GoldReductionToolWithFit):
    """Dimensionality reduction of 2D vectors using UMAP, PCA, TSNE, or GaussianRandomProjection.

    Attributes:
        _reducer: An instance of UMAP, PCA, TSNE, or GaussianRandomProjection.
    """

    def __init__(self, reducer: UMAP | PCA | TSNE | GaussianRandomProjection):
        """Initialize the GoldSKLearnReductionTool.

        Args:
            reducer: An instance of UMAP, PCA, TSNE, or GaussianRandomProjection for dimensionality reduction.
        """
        self._reducer = reducer
        self._has_transform = hasattr(reducer, "transform")

    def fit(self, x: torch.Tensor) -> None:
        """Fit the dimensionality reduction model to the 2D vectors.

        Args:
            x: Already vectorized 2D input tensor of shape (num_vectors, feature_dim) to fit the model on.

        Raises:
            ValueError: If input is not a 2D tensor.
        """
        self._validate_input(x)
        x_np = x.detach().cpu().numpy()
        self._reducer.fit(x_np)

    def fit_transform(self, x: torch.Tensor) -> torch.Tensor:
        """Fit the dimensionality reduction model to the 2D vectors and transform it.

        Args:
            x: Already vectorized 2D input tensor of shape (num_vectors, feature_dim) to fit and transform.

        Returns:
            Transformed tensor with reduced dimensionality.

        Raises:
            ValueError: If input is not a 2D tensor.
        """
        self._validate_input(x)
        x_np = x.detach().cpu().numpy()
        transformed = self._reducer.fit_transform(x_np)
        return torch.from_numpy(transformed).to(device=x.device, dtype=x.dtype)

    def transform(self, x: torch.Tensor) -> torch.Tensor:
        """Transform the 2D vectors using the fitted dimensionality reduction model.

        Args:
            x: Already vectorized 2D input tensor of shape (num_vectors, feature_dim) to transform.

        Returns:
            Transformed tensor with reduced dimensionality.

        Raises:
            ValueError: If input is not a 2D tensor.
        """
        if not self._has_transform:
            logger.warning(
                "The provided reducer does not have a 'transform' method. "
                "Falling back to 'fit_transform' for transformation."
            )
            return self.fit_transform(x)

        self._validate_input(x)
        x_np = x.detach().cpu().numpy()
        transformed = self._reducer.transform(x_np)
        return torch.from_numpy(transformed).to(device=x.device, dtype=x.dtype)



import pytest
import torch
from torch.nn import Linear

from goldener.reduce import GoldSKLearnReductionTool, GoldTorchModuleReductionTool
from umap import UMAP
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from sklearn.random_projection import GaussianRandomProjection


class TestGoldSKLearnReductionTool:
    def test_reducer_rejects_invalid(self):
        """Test that reducer rejects 1D tensor input and requires 2D tensor."""
        data_1d = torch.randn(10, dtype=torch.float32)
        reducer = PCA(n_components=3)
        dr = GoldSKLearnReductionTool(reducer)

        with pytest.raises(
            ValueError, match="GoldReductionTool only accepts 2D tensors"
        ):
            dr.fit(data_1d)

        with pytest.raises(
            ValueError, match="GoldReductionTool only accepts 2D tensors"
        ):
            dr.fit_transform(data_1d)

        dr.fit(torch.randn(10, 5))
        with pytest.raises(
            ValueError, match="GoldReductionTool only accepts 2D tensors"
        ):
            dr.transform(data_1d)

    def test_pca(self):
        data = torch.randn(10, 5, dtype=torch.float32)
        reducer = PCA(n_components=3)
        rd = GoldSKLearnReductionTool(reducer)
        rd.fit(data)
        out = rd.transform(data)
        assert isinstance(out, torch.Tensor)
        assert out.shape == (data.shape[0], 3)
        out2 = rd.fit_transform(data)
        assert isinstance(out2, torch.Tensor)
        assert out2.shape == (data.shape[0], 3)

    def test_tsne(self):
        data = torch.randn(10, 5, dtype=torch.float32)
        reducer = TSNE(n_components=2, perplexity=3)
        rd = GoldSKLearnReductionTool(reducer)
        rd.fit(data)
        out = rd.transform(data)
        assert isinstance(out, torch.Tensor)
        assert out.shape == (data.shape[0], 2)

        out = rd.fit_transform(data)
        assert isinstance(out, torch.Tensor)
        assert out.shape == (data.shape[0], 2)

    def test_umap(self):
        data = torch.randn(10, 5, dtype=torch.float32)
        reducer = UMAP(n_components=2)
        rd = GoldSKLearnReductionTool(reducer)
        rd.fit(data)
        out = rd.transform(data)
        assert isinstance(out, torch.Tensor)
        assert out.shape == (data.shape[0], 2)
        out2 = rd.fit_transform(data)
        assert isinstance(out2, torch.Tensor)
        assert out2.shape == (data.shape[0], 2)

    def test_gaussian_random_projection(self):
        data = torch.randn(10, 5, dtype=torch.float32)
        reducer = GaussianRandomProjection(n_components=4)
        rd = GoldSKLearnReductionTool(reducer)
        rd.fit(data)
        out = rd.transform(data)
        assert isinstance(out, torch.Tensor)
        assert out.shape == (data.shape[0], 4)
        out2 = rd.fit_transform(data)
        assert isinstance(out2, torch.Tensor)
        assert out2.shape == (data.shape[0], 4)


class TestGoldTorchModuleReductionTool:
    def test_reducer_rejects_invalid(self):
        data_1d = torch.randn(10, dtype=torch.float32)
        reducer = Linear(in_features=5, out_features=3)
        dr = GoldTorchModuleReductionTool(reducer)

        with pytest.raises(
            ValueError, match="GoldReductionTool only accepts 2D tensors"
        ):
            dr.transform(data_1d)

    def test_torch_module_reduction_linear(self):
        data = torch.randn(10, 5, dtype=torch.float32)
        linear = Linear(in_features=5, out_features=3)
        rd = GoldTorchModuleReductionTool(linear)
        out = rd.transform(data)
        assert isinstance(out, torch.Tensor)
        assert out.shape == (data.shape[0], 3)
        assert out.dtype == data.dtype



from goldener.clusterize import (
    GoldClusteringTool,
    GoldSKLearnClusteringTool,
    GoldRandomClusteringTool,
    GoldClusterizer,
)
from goldener.describe import (
    GoldDescriptor,
)
from goldener.embed import (
    EmbeddingFusionStrategy,
    GoldEmbeddingFusionTool,
    GoldEmbeddingTool,
    TorchGoldEmbeddingTool,
    TorchGoldEmbeddingToolConfig,
    MultiModalTorchGoldEmbeddingTool,
)
from goldener.pxt_utils import GoldPxtTorchDataset
from goldener.reduce import (
    GoldReductionTool,
    GoldReductionToolWithFit,
    GoldSKLearnReductionTool,
    GoldTorchModuleReductionTool,
)
from goldener.select import (
    GoldSelectionTool,
    GoldSelector,
    GoldGreedyClosestPointSelectionTool,
    GoldGreedyFarthestPointSelectionTool,
    GoldGreedyKCenterSelectionTool,
    GoldGreedyKernelPointsSelectionTool,
)
from goldener.split import GoldSet, GoldSplitter
from goldener.torch_utils import ResetableTorchIterableDataset
from goldener.vectorize import (
    Filter2DWithCount,
    FilterLocation,
    Vectorized,
    TensorVectorizer,
)


__all__ = (
    "GoldClusteringTool",
    "GoldSKLearnClusteringTool",
    "GoldRandomClusteringTool",
    "GoldClusterizer",
    "GoldDescriptor",
    "EmbeddingFusionStrategy",
    "GoldEmbeddingFusionTool",
    "GoldEmbeddingTool",
    "TorchGoldEmbeddingTool",
    "TorchGoldEmbeddingToolConfig",
    "MultiModalTorchGoldEmbeddingTool",
    "GoldPxtTorchDataset",
    "GoldReductionTool",
    "GoldReductionToolWithFit",
    "GoldSKLearnReductionTool",
    "GoldTorchModuleReductionTool",
    "GoldSelectionTool",
    "GoldSelector",
    "GoldGreedyClosestPointSelectionTool",
    "GoldGreedyFarthestPointSelectionTool",
    "GoldGreedyKCenterSelectionTool",
    "GoldGreedyKernelPointsSelectionTool",
    "GoldSet",
    "GoldSplitter",
    "ResetableTorchIterableDataset",
    "Filter2DWithCount",
    "FilterLocation",
    "Vectorized",
    "TensorVectorizer",
)



import numpy as np
import pytest
import torch

import pixeltable as pxt
from torch.utils.data import Dataset

from goldener.clusterize import GoldClusterizer, GoldRandomClusteringTool
from goldener.describe import GoldDescriptor
from goldener.embed import TorchGoldEmbeddingToolConfig, TorchGoldEmbeddingTool
from goldener.pxt_utils import pxt_torch_dataset_collate_fn
from goldener.split import GoldSplitter, GoldSet, check_sets_validity
from goldener.vectorize import TensorVectorizer, GoldVectorizer
from goldener.select import GoldSelector


class DummyModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = torch.nn.Conv2d(3, 4, 3, padding=1)
        self.relu = torch.nn.ReLU()

    def forward(self, x):
        x = self.conv1(x)
        x = self.relu(x)
        return x


@pytest.fixture(scope="function")
def embedder():
    model = DummyModel()
    config = TorchGoldEmbeddingToolConfig(model=model, layers=None)
    return TorchGoldEmbeddingTool(config)


class DummyDataset(Dataset):
    def __init__(self, samples):
        self._samples = samples

    def __len__(self):
        return len(self._samples)

    def __getitem__(self, idx):
        return self._samples[idx]


@pytest.fixture(scope="function")
def descriptor(embedder):
    return GoldDescriptor(
        table_path="unit_test.descriptor_split",
        embedder=embedder,
        to_keep_schema={"label": pxt.String},
        batch_size=2,
        collate_fn=None,
        device=torch.device("cpu"),
        allow_existing=False,
    )


@pytest.fixture(scope="function")
def selector():
    return GoldSelector(
        table_path="unit_test.selector_split",
        to_keep_schema={"label": pxt.String},
        vectorized_key="vectorized",
        batch_size=2,
    )


@pytest.fixture(scope="function")
def vectorizer():
    return GoldVectorizer(
        table_path="unit_test.vectorizer_split",
        vectorizer=TensorVectorizer(),
        to_keep_schema={"label": pxt.String},
        collate_fn=pxt_torch_dataset_collate_fn,
        batch_size=2,
    )


@pytest.fixture(scope="function")
def basic_splitter(descriptor, vectorizer, selector):
    sets = [GoldSet(name="train", size=0.5), GoldSet(name="val", size=0.5)]
    return GoldSplitter(
        sets=sets, descriptor=descriptor, selector=selector, vectorizer=vectorizer
    )


class TestGoldSplitter:
    def test_split_in_table_from_dataset(self, basic_splitter):
        pxt.drop_dir("unit_test", force=True)

        split_table = basic_splitter.split_in_table(
            to_split=DummyDataset(
                [
                    {"data": torch.rand(3, 8, 8), "idx": idx, "label": "dummy"}
                    for idx in range(10)
                ]
            )
        )

        splitted = basic_splitter.get_split_indices(
            split_table,
            selection_key=basic_splitter.selector.selection_key,
            idx_key="idx",
        )

        assert len(splitted) == 2
        assert len(splitted["train"]) == 5
        assert len(splitted["val"]) == 5

        pxt.get_table(basic_splitter.descriptor.table_path)
        pxt.get_table(basic_splitter.vectorizer.table_path)
        pxt.get_table(basic_splitter.selector.table_path)

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_split_in_table_from_table(self, basic_splitter):
        pxt.drop_dir("unit_test", force=True)

        src_path = "unit_test.test_split_in_table"
        pxt.create_dir("unit_test", if_exists="ignore")
        src_table = pxt.create_table(
            src_path,
            source=[
                {
                    "data": torch.rand(3, 8, 8).numpy().astype(np.float32),
                    "idx": idx,
                    "label": "dummy",
                }
                for idx in range(10)
            ],
            if_exists="replace_force",
        )

        split_table = basic_splitter.split_in_table(to_split=src_table)

        splitted = basic_splitter.get_split_indices(
            split_table,
            selection_key=basic_splitter.selector.selection_key,
            idx_key="idx",
        )

        assert len(splitted) == 2
        assert len(splitted["train"]) == 5
        assert len(splitted["val"]) == 5

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_split_with_label(self, descriptor, selector, vectorizer):
        pxt.drop_dir("unit_test", force=True)

        sets = [
            GoldSet(name="train", size=0.5),
            GoldSet(name="val", size=0.5),
        ]
        selector.label_key = "label"
        splitter = GoldSplitter(
            sets=sets,
            descriptor=descriptor,
            selector=selector,
            vectorizer=vectorizer,
        )

        split_table = splitter.split_in_table(
            to_split=DummyDataset(
                [
                    {"data": torch.rand(3, 8, 8), "idx": idx, "label": str(idx % 2)}
                    for idx in range(10)
                ]
            )
        )
        splitted = splitter.get_split_indices(
            split_table,
            selection_key=splitter.selector.selection_key,
            idx_key="idx",
        )

        assert set(splitted.keys()) == {"train", "val"}
        assert len(splitted["train"]) == 5
        assert len(splitted["val"]) == 5

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_split_with_excluded_labels(self, descriptor, vectorizer, embedder):
        pxt.drop_dir("unit_test", force=True)

        sets = [
            GoldSet(name="train", size=0.5),
            GoldSet(name="val", size=0.5),
        ]
        selector = GoldSelector(
            table_path="unit_test.selector_split",
            to_keep_schema={"label": pxt.String},
            vectorized_key="vectorized",
            batch_size=2,
            label_key="label",
            exclude_labels={"excluded"},
        )
        splitter = GoldSplitter(
            sets=sets,
            descriptor=descriptor,
            selector=selector,
            vectorizer=vectorizer,
        )

        split_table = splitter.split_in_table(
            to_split=DummyDataset(
                [
                    {
                        "data": torch.rand(3, 8, 8),
                        "idx": idx,
                        "label": "excluded" if idx < 4 else str(idx % 2),
                    }
                    for idx in range(10)
                ]
            )
        )

        for row in split_table.collect():
            assert row["label"] != "excluded"

        splitted = splitter.get_split_indices(
            split_table,
            selection_key=splitter.selector.selection_key,
            idx_key="idx",
        )

        assert set(splitted.keys()) == {"train", "val"}
        # Only 6 samples (idx 4-9) remain after excluding idx 0-3
        assert len(splitted["train"]) + len(splitted["val"]) == 6

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_duplicated_set_names(self, descriptor, selector, vectorizer):
        pxt.drop_dir("unit_test", force=True)

        sets = [GoldSet(name="train", size=0.5), GoldSet(name="train", size=0.5)]
        with pytest.raises(ValueError, match="Set names must be unique"):
            GoldSplitter(
                sets=sets,
                descriptor=descriptor,
                selector=selector,
                vectorizer=vectorizer,
            )

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_not_complete_set_sizes_as_float(self, descriptor, selector, vectorizer):
        pxt.drop_dir("unit_test", force=True)

        with pytest.raises(
            ValueError, match="Sampling size as float must be equal to 1.0"
        ):
            GoldSplitter(
                sets=[GoldSet(name="train", size=0.5), GoldSet(name="val", size=0.4)],
                descriptor=descriptor,
                selector=selector,
                vectorizer=vectorizer,
            )

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_not_complete_set_sizes_as_int(self, descriptor, selector, vectorizer):
        pxt.drop_dir("unit_test", force=True)

        gold_splitter = GoldSplitter(
            sets=[GoldSet(name="train", size=2), GoldSet(name="val", size=2)],
            descriptor=descriptor,
            selector=selector,
            vectorizer=vectorizer,
        )

        with pytest.raises(
            ValueError,
            match="Sampling size as int must be equal to the total number of samples",
        ):
            gold_splitter.split_in_table(
                to_split=DummyDataset(
                    [
                        {"data": torch.rand(3, 8, 8), "idx": idx, "label": "dummy"}
                        for idx in range(10)
                    ]
                )
            )

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_different_type_set_size(self, descriptor, selector, vectorizer):
        pxt.drop_dir("unit_test", force=True)

        with pytest.raises(ValueError, match="All set sizes must be of the same type"):
            GoldSplitter(
                sets=[GoldSet(name="train", size=0.5), GoldSet(name="val", size=3)],
                descriptor=descriptor,
                selector=selector,
                vectorizer=vectorizer,
            )

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_label_key_not_found(self, descriptor, selector, vectorizer):
        pxt.drop_dir("unit_test", force=True)
        selector.label_key = "nonexistent"
        sets = [
            GoldSet(name="train", size=0.5),
            GoldSet(name="val", size=0.5),
        ]
        splitter = GoldSplitter(
            sets=sets,
            descriptor=descriptor,
            selector=selector,
            vectorizer=vectorizer,
        )

        with pytest.raises(
            ValueError, match="label_key and label_value must be set together"
        ):
            splitter.split_in_table(
                to_split=DummyDataset(
                    [
                        {"data": torch.rand(3, 8, 8), "idx": idx, "label": "dummy"}
                        for idx in range(10)
                    ]
                )
            )

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_set_with_small_population(self, descriptor, selector, vectorizer):
        pxt.drop_dir("unit_test", force=True)

        sets = [
            GoldSet(name="train", size=0.001),
            GoldSet(name="val", size=0.999),
        ]
        splitter = GoldSplitter(
            sets=sets, descriptor=descriptor, selector=selector, vectorizer=vectorizer
        )

        split_table = splitter.split_in_table(
            to_split=DummyDataset(
                [
                    {"data": torch.rand(3, 8, 8), "idx": idx, "label": "dummy"}
                    for idx in range(2)
                ]
            )
        )

        splitted = splitter.get_split_indices(
            split_table,
            selection_key=splitter.selector.selection_key,
            idx_key="idx",
        )

        assert set(splitted.keys()) == {"train", "val"}
        assert len(splitted["train"]) == 1

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_class_with_small_population_failure(
        self, descriptor, selector, vectorizer
    ):
        pxt.drop_dir("unit_test", force=True)
        selector.label_key = "label"
        sets = [
            GoldSet(name="train", size=0.0001),
            GoldSet(name="val", size=0.9999),
        ]
        splitter = GoldSplitter(
            sets=sets,
            descriptor=descriptor,
            selector=selector,
            vectorizer=vectorizer,
        )

        with pytest.raises(ValueError, match="While trying to adjust counts"):
            splitter.split_in_table(
                to_split=DummyDataset(
                    [
                        {
                            "data": torch.rand(3, 8, 8),
                            "idx": idx,
                            "label": "A" if idx < 5 else "B",
                        }
                        for idx in range(10)
                    ]
                )
            )

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_max_batches(self, descriptor, selector, vectorizer):
        pxt.drop_dir("unit_test", force=True)

        sets = [GoldSet(name="train", size=0.5), GoldSet(name="val", size=0.5)]
        splitter = GoldSplitter(
            sets=sets,
            descriptor=descriptor,
            selector=selector,
            vectorizer=vectorizer,
            max_batches=1,
        )

        assert splitter.descriptor.max_batches == 1

        split_table = splitter.split_in_table(
            to_split=DummyDataset(
                [
                    {"data": torch.rand(3, 8, 8), "idx": idx, "label": "dummy"}
                    for idx in range(10)
                ]
            )
        )
        splitted = splitter.get_split_indices(
            split_table,
            selection_key=splitter.selector.selection_key,
            idx_key="idx",
        )

        assert len(splitted) == 2
        # Only 2 items total (1 batch with batch_size=2)
        assert len(splitted["train"]) + len(splitted["val"]) == 2

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_with_no_remaining_indices(self, descriptor, selector, vectorizer):
        pxt.drop_dir("unit_test", force=True)

        sets = [GoldSet(name="train", size=0.5), GoldSet(name="val", size=0.5)]
        selector.max_batches = 1
        vectorizer.max_batches = 1
        splitter = GoldSplitter(
            sets=sets,
            descriptor=descriptor,
            selector=selector,
            vectorizer=vectorizer,
            max_batches=1,
        )

        with pytest.raises(
            ValueError,
            match="Not enough data to split among",
        ):
            splitter.split_in_table(
                to_split=DummyDataset(
                    [
                        {"data": torch.rand(3, 8, 8), "idx": idx, "label": "dummy"}
                        for idx in range(10)
                    ]
                )
            )

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_split_in_dataset_from_dataset(self, basic_splitter):
        pxt.drop_dir("unit_test", force=True)
        basic_splitter.in_described_table = True
        basic_splitter.drop_table = True
        split_dataset = basic_splitter.split_in_dataset(
            to_split=DummyDataset(
                [
                    {"data": torch.rand(3, 8, 8), "idx": idx, "label": "dummy"}
                    for idx in range(10)
                ]
            )
        )

        train_count = 0
        val_count = 0
        for item in split_dataset:
            if item["selected"] == "train":
                train_count += 1
            elif item["selected"] == "val":
                val_count += 1
            else:
                raise ValueError("Unknown split name found in dataset.")

        assert train_count == 5
        assert val_count == 5

        assert (
            pxt.get_table(basic_splitter.descriptor.table_path, if_not_exists="ignore")
            is None
        )
        assert (
            pxt.get_table(basic_splitter.vectorizer.table_path, if_not_exists="ignore")
            is None
        )
        assert (
            pxt.get_table(basic_splitter.selector.table_path, if_not_exists="ignore")
            is None
        )

        split_dataset.keep_cache = False

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_split_in_dataset_without_drop(self, basic_splitter):
        pxt.drop_dir("unit_test", force=True)
        basic_splitter.in_described_table = True
        basic_splitter.drop_table = False
        split_dataset = basic_splitter.split_in_dataset(
            to_split=DummyDataset(
                [
                    {"data": torch.rand(3, 8, 8), "idx": idx, "label": "dummy"}
                    for idx in range(10)
                ]
            )
        )

        assert (
            pxt.get_table(basic_splitter.descriptor.table_path, if_not_exists="ignore")
            is not None
        )
        assert (
            pxt.get_table(basic_splitter.vectorizer.table_path, if_not_exists="ignore")
            is not None
        )
        assert (
            pxt.get_table(basic_splitter.selector.table_path, if_not_exists="ignore")
            is not None
        )

        split_dataset.keep_cache = False

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_split_in_table_from_dataset_with_drop_table(self, basic_splitter):
        pxt.drop_dir("unit_test", force=True)

        basic_splitter.drop_table = True
        basic_splitter.split_in_table(
            to_split=DummyDataset(
                [
                    {"data": torch.rand(3, 8, 8), "idx": idx, "label": "dummy"}
                    for idx in range(10)
                ]
            )
        )

        with pytest.raises(pxt.Error):
            pxt.get_table(basic_splitter.descriptor.table_path)

        with pytest.raises(pxt.Error):
            pxt.get_table(basic_splitter.vectorizer.table_path)

        basic_splitter.in_described_table = True
        basic_splitter.split_in_table(
            to_split=DummyDataset(
                [
                    {"data": torch.rand(3, 8, 8), "idx": idx, "label": "dummy"}
                    for idx in range(10)
                ]
            )
        )

        with pytest.raises(pxt.Error):
            pxt.get_table(basic_splitter.selector.table_path)

        with pytest.raises(pxt.Error):
            pxt.get_table(basic_splitter.vectorizer.table_path)

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_split_in_table_from_dataset_with_restart(self, basic_splitter):
        pxt.drop_dir("unit_test", force=True)

        dataset = DummyDataset(
            [
                {"data": torch.rand(3, 8, 8), "idx": idx, "label": "dummy"}
                for idx in range(10)
            ]
        )
        basic_splitter.max_batches = 2
        basic_splitter.split_in_table(to_split=dataset)

        basic_splitter.max_batches = None
        split_table = basic_splitter.split_in_table(to_split=dataset)
        splitted = basic_splitter.get_split_indices(
            split_table,
            selection_key=basic_splitter.selector.selection_key,
            idx_key="idx",
        )

        assert len(splitted) == 2
        assert len(splitted["train"]) == 5
        assert len(splitted["val"]) == 5

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_get_split_indices_from_table(self):
        pxt.drop_dir("unit_test", force=True)

        src_path = "unit_test.test_split_in_table"
        pxt.create_dir("unit_test", if_exists="ignore")
        split_table = pxt.create_table(
            src_path,
            source=[
                {
                    "data": torch.rand(3, 8, 8).numpy().astype(np.float32),
                    "idx": idx,
                    "label": "dummy",
                    "set": "train" if idx < 5 else "val",
                }
                for idx in range(10)
            ],
            if_exists="replace_force",
        )

        splitted = GoldSplitter.get_split_indices(
            split_table,
            selection_key="set",
            idx_key="idx",
        )

        for set_name, indices in splitted.items():
            assert len(indices) == 5
            assert set_name in ["train", "val"]

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_get_split_indices_from_dataset(self):
        pxt.drop_dir("unit_test", force=True)

        split_dataset = DummyDataset(
            [
                {
                    "data": torch.rand(3, 8, 8),
                    "idx": idx,
                    "set": "train" if idx < 5 else "val",
                    "label": "dummy",
                }
                for idx in range(10)
            ]
        )

        splitted = GoldSplitter.get_split_indices(
            split_dataset,
            selection_key="set",
            idx_key="idx",
        )

        for set_name, indices in splitted.items():
            assert len(indices) == 5
            assert set_name in ["train", "val"]

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_with_descriptor_including_vectorizer(self, embedder):
        pxt.drop_dir("unit_test", force=True)

        descriptor = GoldDescriptor(
            table_path="unit_test.descriptor_split",
            embedder=embedder,
            vectorizer=TensorVectorizer(),
            description_key="vectorized",
            batch_size=2,
            collate_fn=None,
            device=torch.device("cpu"),
            allow_existing=False,
        )

        selector = GoldSelector(
            table_path="unit_test.selector_split",
            vectorized_key="vectorized",
            batch_size=2,
        )

        sets = [GoldSet(name="train", size=0.5), GoldSet(name="val", size=0.5)]
        splitter = GoldSplitter(
            sets=sets,
            descriptor=descriptor,
            selector=selector,
            vectorizer=None,
            max_batches=1,
        )

        assert splitter.descriptor.max_batches == 1

        split_table = splitter.split_in_table(
            to_split=DummyDataset(
                [
                    {"data": torch.rand(3, 8, 8), "idx": idx, "label": "dummy"}
                    for idx in range(10)
                ]
            )
        )
        splitted = splitter.get_split_indices(
            split_table,
            selection_key=splitter.selector.selection_key,
            idx_key="idx",
        )

        assert len(splitted) == 2
        # Only 2 items total (1 batch with batch_size=2)
        assert len(splitted["train"]) + len(splitted["val"]) == 2

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_without_descriptor(self, basic_splitter):
        pxt.drop_dir("unit_test", force=True)

        basic_splitter.descriptor = None
        basic_splitter.vectorizer.collate_fn = None
        split_table = basic_splitter.split_in_table(
            to_split=DummyDataset(
                [
                    {"embeddings": torch.rand(4, 8, 8), "idx": idx, "label": "dummy"}
                    for idx in range(10)
                ]
            )
        )
        splitted = basic_splitter.get_split_indices(
            split_table,
            selection_key=basic_splitter.selector.selection_key,
            idx_key="idx",
        )

        assert len(splitted) == 2
        # Only 2 items total (1 batch with batch_size=2)
        assert len(splitted["train"]) + len(splitted["val"]) == 10

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_without_descriptor_nor_vectorizer(self, basic_splitter):
        pxt.drop_dir("unit_test", force=True)

        basic_splitter.descriptor = None
        basic_splitter.vectorizer = None
        split_table = basic_splitter.split_in_table(
            to_split=DummyDataset(
                [
                    {
                        "vectorized": torch.rand(
                            4,
                        ),
                        "idx": idx,
                        "label": "dummy",
                    }
                    for idx in range(10)
                ]
            )
        )
        splitted = basic_splitter.get_split_indices(
            split_table,
            selection_key=basic_splitter.selector.selection_key,
            idx_key="idx",
        )

        assert len(splitted) == 2
        # Only 2 items total (1 batch with batch_size=2)
        assert len(splitted["train"]) + len(splitted["val"]) == 10

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_with_int_sizes(self, basic_splitter):
        pxt.drop_dir("unit_test", force=True)

        basic_splitter.descriptor = None
        basic_splitter.vectorizer = None
        basic_splitter.sets = [
            GoldSet(name="train", size=6),
            GoldSet(name="val", size=4),
        ]
        split_table = basic_splitter.split_in_table(
            to_split=DummyDataset(
                [
                    {
                        "vectorized": torch.rand(
                            4,
                        ),
                        "idx": idx,
                        "label": "dummy",
                    }
                    for idx in range(10)
                ]
            )
        )
        splitted = basic_splitter.get_split_indices(
            split_table,
            selection_key=basic_splitter.selector.selection_key,
            idx_key="idx",
        )

        assert len(splitted) == 2
        assert len(splitted["train"]) + len(splitted["val"]) == 10

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_with_no_descriptor_but_in_described_table(self, selector):
        pxt.drop_dir("unit_test", force=True)

        with pytest.raises(
            ValueError,
            match="in_described_table is set to True, but no descriptor is provided",
        ):
            GoldSplitter(
                sets=[GoldSet(name="train", size=0.5), GoldSet(name="val", size=0.5)],
                descriptor=None,
                selector=selector,
                vectorizer=None,
                in_described_table=True,
            )

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_with_vectorizer_but_wrong_selection_key(self, selector, vectorizer):
        pxt.drop_dir("unit_test", force=True)

        vectorizer.vectorized_key = "wrong_key"
        with pytest.raises(
            ValueError, match="does not match selector's vectorized_key"
        ):
            GoldSplitter(
                sets=[GoldSet(name="train", size=0.5), GoldSet(name="val", size=0.5)],
                descriptor=None,
                selector=selector,
                vectorizer=vectorizer,
                in_described_table=False,
            )

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_with_descriptor_but_wrong_selection_key(self, selector, descriptor):
        pxt.drop_dir("unit_test", force=True)

        descriptor.description_key = "wrong_key"
        with pytest.raises(
            ValueError, match="does not match selector's vectorized_key"
        ):
            GoldSplitter(
                sets=[GoldSet(name="train", size=0.5), GoldSet(name="val", size=0.5)],
                descriptor=descriptor,
                selector=selector,
                vectorizer=None,
                in_described_table=False,
            )

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_with_descriptor_and_vectorizer_but_wrong_key(
        self, selector, descriptor, vectorizer
    ):
        pxt.drop_dir("unit_test", force=True)

        descriptor.description_key = "wrong_key"
        with pytest.raises(ValueError, match="does not match vectorizer's data_key"):
            GoldSplitter(
                sets=[GoldSet(name="train", size=0.5), GoldSet(name="val", size=0.5)],
                descriptor=descriptor,
                selector=selector,
                vectorizer=vectorizer,
                in_described_table=False,
            )

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)

    def test_split_with_clusterizer(self, selector):
        pxt.drop_dir("unit_test", force=True)

        # Build a simple vectorized table first, then let the splitter handle clustering
        sets = [GoldSet(name="train", size=0.5), GoldSet(name="val", size=0.5)]

        clusterizer = GoldClusterizer(
            table_path="unit_test.clusterizer_split",
            clustering_tool=GoldRandomClusteringTool(random_state=0),
            label_key="label",
            allow_existing=False,
        )

        splitter = GoldSplitter(
            sets=sets,
            descriptor=None,
            selector=selector,
            vectorizer=None,
            clusterizer=clusterizer,
            n_clusters=2,
        )

        split_table = splitter.split_in_table(
            to_split=DummyDataset(
                [
                    {
                        "vectorized": torch.rand(
                            3,
                        ),
                        "idx": idx_vector // 4,
                        "label": "dummy",
                        "idx_vector": idx_vector,
                    }
                    for idx_vector in range(120)
                ]
            )
        )

        splitted = splitter.get_split_indices(
            split_table,
            selection_key=splitter.selector.selection_key,
            idx_key="idx",
        )

        assert set(splitted.keys()) == {"train", "val"}
        assert len(splitted["train"]) == 15
        assert len(splitted["val"]) == 15

        # Check that the clusterizer table was created and has clusters assigned
        cluster_table = pxt.get_table(clusterizer.table_path)
        assert cluster_table.count() > 0
        distinct_clusters = [
            row[clusterizer.cluster_key]
            for row in cluster_table.select(cluster_table[clusterizer.cluster_key])
            .distinct()
            .collect()
        ]
        # Since we requested 2 clusters, resulting cluster ids should be a subset of {0,1}
        assert set(distinct_clusters).issubset({0, 1})

        pxt.drop_dir("unit_test", if_not_exists="ignore", force=True)


class TestCheckSetsValidity:
    def test_sum_of_sizes_less_than_one(self):
        sets = [GoldSet(name="train", size=0.4), GoldSet(name="val", size=0.5)]
        check_sets_validity(sets, force_max=False)

    def test_sum_of_sizes_equal_one(self):
        sets = [GoldSet(name="train", size=0.5), GoldSet(name="val", size=0.5)]
        check_sets_validity(sets, force_max=True)

    def test_with_float_failure(self):
        with pytest.raises(
            ValueError,
            match="Sampling size as float must be greater than 0.0 and at most 1.0",
        ):
            sets = [GoldSet(name="train", size=0.51), GoldSet(name="val", size=0.5)]
            check_sets_validity(sets, force_max=False)

        with pytest.raises(
            ValueError,
            match="Sampling size as float must be equal to 1.0",
        ):
            sets = [GoldSet(name="train", size=0.49), GoldSet(name="val", size=0.5)]
            check_sets_validity(sets, force_max=True)

    def test_sum_of_sizes_less_than_total(self):
        sets = [GoldSet(name="train", size=2), GoldSet(name="val", size=2)]
        check_sets_validity(sets, total=5, force_max=False)

    def test_sum_of_sizes_equal_total(self):
        sets = [GoldSet(name="train", size=2), GoldSet(name="val", size=3)]
        check_sets_validity(sets, total=5, force_max=True)

    def test_with_int_failure(self):
        with pytest.raises(
            ValueError,
            match="Sampling size as int must be equal to the total number of samples",
        ):
            sets = [GoldSet(name="train", size=3), GoldSet(name="val", size=1)]
            check_sets_validity(sets, total=5, force_max=True)

        with pytest.raises(
            ValueError,
            match="Sampling size as int must be greater than 0 and less or equal than the total number of samples",
        ):
            sets = [GoldSet(name="train", size=3), GoldSet(name="val", size=3)]
            check_sets_validity(sets, total=5, force_max=False)

    def test_same_name(self):
        sets = [GoldSet(name="train", size=3), GoldSet(name="train", size=2)]
        with pytest.raises(ValueError, match="Set names must be unique"):
            check_sets_validity(sets, total=5)



from abc import ABC, abstractmethod
from logging import getLogger
from typing import Callable, Any

import jax
import numpy as np
import pixeltable as pxt
from pixeltable import Error, Query
from pixeltable.catalog import Table

import torch
import jax.numpy as jnp

from coreax import SupervisedData, ScalarValuedKernel
from coreax.kernels import LinearKernel
from coreax.solvers import GreedyKernelPoints
from torch.utils.data import Dataset, DataLoader
from tqdm import tqdm

from goldener.clusterize import get_random_chunk_assignment
from goldener.pxt_utils import (
    set_value_to_idx_rows,
    GoldPxtTorchDataset,
    get_expr_from_column_name,
    get_valid_table,
    make_batch_ready_for_table,
    check_pxt_table_has_primary_key,
    get_sample_row_from_idx,
)
from goldener.reduce import GoldReductionTool, GoldReductionToolWithFit
from goldener.torch_utils import get_dataset_sample_dict
from goldener.utils import (
    filter_batch_from_indices,
    get_indices_with_labels,
    split_sampling_among_chunks,
    get_sampling_count_from_ratios,
    get_ratios_for_counts,
)

logger = getLogger(__name__)


class GoldSelectionTool(ABC):
    """Run selection of a subset of data points from vectorized samples using a coresubset selection algorithm."""

    def _validate_input(self, x: torch.Tensor) -> None:
        """Validate that input is already vectorized (2D torch.Tensor).

        Args:
            x: Input tensor to validate.

        Raises:
            ValueError: If input is not a 2D tensor.
        """
        if x.ndim != 2:
            raise ValueError(
                f"GoldSelectionTool only accepts 2D tensors (num_vectors, feature_dim). "
                f"Got shape {x.shape}. Please ensure your input is already vectorized."
            )

    @abstractmethod
    def select(
        self,
        x: torch.Tensor,
        k: int,
    ) -> list[int]:
        """Select a subset of data points from the input data.

        Args:
            x: A 2D torch.Tensor where each row represents a data point.
            k: The number of data points to select.

        Returns:
            A list of indices corresponding to the selected data points.
        """


class GoldGreedyKernelPointsSelectionTool(GoldSelectionTool):
    """Coresubset selection using Coreax's GreedyKernelPoints solver.

    The solver is created at each request depending on the specified k points to select.

    Attributes:
        feature_kernel: The kernel to use for the GreedyKernelPoints solver.
        random_key: Random key for reproducibility in the GreedyKernelPoints solver.
    """

    def __init__(
        self,
        feature_kernel: ScalarValuedKernel,
        random_key: int = 42,
    ) -> None:
        self.random_key = random_key
        self.feature_kernel = feature_kernel

    def select(
        self,
        x: torch.Tensor,
        k: int,
    ) -> list[int]:
        """Select a subset of data points from the input data using greedy kernel point coresubset sampling.

        Args:
            x: A 2D torch.Tensor where each row represents a data point.
            k: The number of data points to select.

        Returns:
            A list of indices corresponding to the selected data points.
        """
        self._validate_input(x)
        in_data = SupervisedData(
            data=jnp.array(x.numpy(force=True)), supervision=jnp.array(np.ones(len(x)))
        )
        solver: GreedyKernelPoints = GreedyKernelPoints(
            coreset_size=k,
            random_key=jax.random.key(self.random_key),
            feature_kernel=self.feature_kernel,
        )
        coreset, _ = solver.reduce(in_data)

        return coreset.unweighted_indices.tolist()


class GoldGreedyKCenterSelectionTool(GoldSelectionTool):
    """Create a coresubset from selecting with the K Center greedy algorithm.

    This is a greedy algorithm that selects iteratively the farthest to the already selected points
    among the not selected points. The first center is selected as the point with
    the largest aggregated distance to all the other points.

    Attributes:
        device: The torch device to use for computations.
    """

    def __init__(self, device: torch.device | str) -> None:
        self.device = device

    def select(
        self,
        x: torch.Tensor,
        k: int,
    ) -> list[int]:
        """Select a subset of data points from the input data by choosing iteratively the point
        with the farthest nearest neighbors among the not selected points.

        Args:
            x: A 2D torch.Tensor where each row represents a data point.
            k: The number of data points to select.

        Returns:
            A list of indices corresponding to the selected data points.
        """
        self._validate_input(x)
        x_len = len(x)
        if k > x_len:
            raise ValueError("k cannot be greater than the number of data points in x.")

        if k == x_len:
            return list(range(x_len))

        x = x.to(self.device)

        # the first point is the one with the largest aggregated distance to all the other points
        sample_to_sample_distance = torch.cdist(x, x)
        first_idx = int(sample_to_sample_distance.norm(dim=1).argmax().item())
        selected_indices = [first_idx]
        if k == 1:
            return selected_indices

        # the next points are selected iteratively as the ones
        # with the largest distance to the already selected points
        distances = sample_to_sample_distance[first_idx]
        for _ in range(k - 1):
            next_idx = int(distances.argmax().item())
            selected_indices.append(next_idx)

            distances_to_next = sample_to_sample_distance[next_idx]
            distances = torch.stack([distances, distances_to_next]).min(dim=0).values

        return selected_indices


class GoldGreedyClosestPointSelectionTool(GoldSelectionTool):
    """Create a coresubset from selecting the closest points iteratively.

    This is a greedy algorithm that selects iteratively the point with the closest nearest neighbors
    among the not selected points.

    Attributes:
        device: The torch device to use for computations.
    """

    def __init__(self, device: torch.device | str) -> None:
        self.device = device

    def select(
        self,
        x: torch.Tensor,
        k: int,
    ) -> list[int]:
        """Select a subset of data points from the input data by choosing iteratively the point
        with the closest nearest neighbors among the not selected points.

        Args:
            x: A 2D torch.Tensor where each row represents a data point.
            k: The number of data points to select.

        Returns:
            A list of indices corresponding to the selected data points.
        """
        self._validate_input(x)
        x_len = len(x)
        if k > x_len:
            raise ValueError("k cannot be greater than the number of data points in x.")

        if k == x_len:
            return list(range(x_len))

        x = x.to(self.device)

        selected_indices: list[int] = []
        remaining_indices = set(range(len(x)))

        for _ in range(k):
            remaining_indices_as_list = list(remaining_indices)
            remaining_vectors = x[remaining_indices_as_list]

            distance = torch.cdist(remaining_vectors, remaining_vectors)
            distance = (
                distance
                + torch.eye(len(remaining_vectors), device=x.device, dtype=x.dtype)
                * distance.max()
            )  # set self-distance to max

            point_with_closest = int(distance.min(dim=1).values.argmin().item())
            index_in_original = remaining_indices_as_list[point_with_closest]

            selected_indices.append(index_in_original)
            remaining_indices.remove(index_in_original)

        return selected_indices


class GoldGreedyFarthestPointSelectionTool(GoldSelectionTool):
    """Create a coresubset from selecting the farthest points iteratively.

    This is a greedy algorithm that selects iteratively the point with the farthest nearest neighbors
    among the not selected points.

    Attributes:
        device: The torch device to use for computations.
    """

    def __init__(self, device: torch.device | str) -> None:
        self.device = device

    def select(
        self,
        x: torch.Tensor,
        k: int,
    ) -> list[int]:
        """Select a subset of data points from the input data by choosing iteratively the point
        with the farthest nearest neighbors among the not selected points.

        Args:
            x: A 2D torch.Tensor where each row represents a data point.
            k: The number of data points to select.

        Returns:
            A list of indices corresponding to the selected data points.
        """
        self._validate_input(x)
        x_len = len(x)
        if k > x_len:
            raise ValueError("k cannot be greater than the number of data points in x.")

        if k == x_len:
            return list(range(x_len))

        x = x.to(self.device)

        selected_indices: list[int] = []
        remaining_indices = set(range(len(x)))

        for _ in range(k):
            remaining_indices_as_list = list(remaining_indices)
            remaining_vectors = x[remaining_indices_as_list]

            distance = torch.cdist(remaining_vectors, remaining_vectors)
            distance = (
                distance
                + torch.eye(len(remaining_vectors), device=x.device, dtype=x.dtype)
                * distance.max()
            )  # set self-distance to max
            point_with_farthest = int(distance.min(dim=1).values.argmax().item())
            index_in_original = remaining_indices_as_list[point_with_farthest]

            selected_indices.append(index_in_original)
            remaining_indices.remove(index_in_original)

        return selected_indices


class GoldSelector:
    """Select a subset of data points from vectorized samples.

    The GoldSelector processes a dataset or PixelTable table to perform coresubset selection using a
    selection algorithm on already vectorized representations. The selection results are stored
    in a local PixelTable table (specified by `table_path`) so that the selection process is idempotent:
    calling the same operation multiple times will not duplicate or recompute selections that are already
    present in the table.

    If the dataset is too big to fit into memory or the coresubset selection algorithm is too
    computationally expensive, the coresubset selection can be performed in chunks.
    All torch tensors will be converted to numpy arrays before saving.

    The selection can operate in a sequential (single-process) mode or a
    distributed mode (not implemented). The table schema is created/validated automatically and will include
    minimal indexing columns (`idx`, `idx_vector`) required to link selected samples back to
    their originating data points.

    Attributes:
        table_path: Path to the PixelTable table where selection results will be stored locally.
        selection_tool: The GoldSelectionTool implementing the selection algorithm.
        reducer: Optional GoldReductionTool instance for dimensionality reduction before selection.
        chunk: Optional chunk size for processing data in chunks to reduce memory consumption.
        collate_fn: Optional function to collate dataset samples into batches composed of
            dictionaries with at least the key specified by `vectorized_key` returning a PyTorch Tensor.
            If None, the dataset is expected to directly provide such batches.
        vectorized_key: Key in the batch dictionary that contains the vectorized data for selection. Default is "vectorized".
        include_vectorized_in_table: Whether to include the vectorized data in the selection table. Defaults to False.
        It is only applied if the cluster table is created from a Table (it is forced anyway for Dataset).
        selection_key: Column name to store the selection value in the PixelTable table. Default is "selected".
        label_key: Optional key for label-based stratified selection.
        exclude_labels: Optional set of label strings to exclude from selection. Default is None.
        to_keep_schema: Optional dictionary defining additional columns to keep from the original dataset/table
            into the selection table. The keys are the column names and the values are the PixelTable types.
        min_pxt_insert_size: Minimum number of rows to accumulate before inserting into PixelTable. Default is 100.
        batch_size: Batch size used when iterating over the data.
        num_workers: Number of workers for the PyTorch DataLoader during iteration on data.
        allow_existing: If False, an error will be raised when the table already exists. Default is True.
        distribute: Whether to use distributed processing for selection and table population. Not implemented yet. Default is False.
        drop_table: Whether to drop the selection table after creating the dataset with selection results. It is only applied
            when using `select_in_dataset`. Default is False.
        max_batches: Optional maximum number of batches to process. Useful for testing on a small subset of the dataset.
        random_state: Optional random state for reproducibility during chunk assignment. Default is None.
    """

    _MINIMAL_SCHEMA: dict[str, type] = {
        "idx": pxt.Required[pxt.Int],
        "idx_vector": pxt.Required[pxt.Int],
    }

    def __init__(
        self,
        table_path: str,
        selection_tool: GoldSelectionTool = GoldGreedyKernelPointsSelectionTool(
            feature_kernel=LinearKernel(output_scale=1, constant=0)
        ),
        reducer: GoldReductionTool | None = None,
        chunk: int | None = None,
        collate_fn: Callable | None = None,
        vectorized_key: str = "vectorized",
        include_vectorized_in_table: bool = False,
        selection_key: str = "selected",
        label_key: str | None = None,
        exclude_labels: set[str] | None = None,
        to_keep_schema: dict[str, type] | None = None,
        min_pxt_insert_size: int = 100,
        batch_size: int = 1,
        num_workers: int = 0,
        allow_existing: bool = True,
        distribute: bool = False,
        drop_table: bool = False,
        max_batches: int | None = None,
        random_state: int | None = None,
    ) -> None:
        """Initialize the GoldSelector.

        Args:
            table_path: Path to store the PixelTable table.
            selection_tool: The GoldSelectionTool implementing the selection algorithm.
            reducer: Optional GoldReductionTool instance for dimensionality reduction before selection.
            chunk: Optional chunk size for processing data in chunks.
            collate_fn: Optional collate function for the DataLoader.
            vectorized_key: Key pointing to the vector for selection. Defaults to "vectorized".
            include_vectorized_in_table: Whether to include the vectorized data in the selection table. Defaults to False.
                It is only applied if the selection table is created from a Table (it is forced anyway for Dataset).
            selection_key: Key for storing selection values. Defaults to "selected".
            label_key: Optional key for label stratification.
            exclude_labels: Optional set of label strings to exclude from selection. Default is None.
            to_keep_schema: Optional schema for additional columns to preserve.
            min_pxt_insert_size: Minimum number of rows to accumulate before inserting into PixelTable. Default is 100.
            batch_size: Batch size used when iterating over the data.
            num_workers: Number of workers for the PyTorch DataLoader during iteration on data.
            allow_existing: Whether to allow using an existing table. Defaults to True.
            distribute: Whether to use distributed selection. Defaults to False.
            drop_table: Whether to drop the table after dataset creation. Defaults to False.
            max_batches: Optional maximum number of batches to process.
            random_state: Optional random state for reproducibility during chunk assignment. Default is None.
        """
        self.table_path = table_path
        self.selection_tool = selection_tool
        self.reducer = reducer
        self.chunk = chunk
        self.collate_fn = collate_fn
        self.vectorized_key = vectorized_key
        self.include_vectorized_in_table = include_vectorized_in_table
        self.selection_key = selection_key
        if exclude_labels is not None and label_key is None:
            raise ValueError(
                "If exclude_labels is provided, label_key must also be provided."
            )
        self.label_key = label_key
        self.exclude_labels = exclude_labels
        self.to_keep_schema = to_keep_schema
        self.min_pxt_insert_size = min_pxt_insert_size
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.allow_existing = allow_existing
        self.distribute = distribute
        self.drop_table = drop_table
        self.max_batches = max_batches
        self.random_state = random_state

    def setup_selection_table(self, select_from: Dataset | Table) -> Table:
        """Set up the selection table based on the source dataset or table.

        Args:
            select_from: Dataset or Table to select from. If a Dataset is provided, each item should be a
                dictionary with at least the `vectorized_key` and `idx` keys after applying the collate_fn.
                If the collate_fn is None, the dataset is expected to directly provide such batches.
                If a Table is provided, it should contain at least the `vectorized_key`,
                `idx` and `idx_vector` columns.

        Returns: A PixelTable Table with the appropriate schema for selection,
            populated with rows corresponding to the source dataset or table.
        """
        logger.info(f"Loading the existing selection table from {self.table_path}")
        try:
            old_selection_table = pxt.get_table(
                self.table_path,
                if_not_exists="ignore",
            )
        except Error:
            logger.info(f"No existing selection table from {self.table_path}")
            old_selection_table = None

        # selection table are expected to have a primary key allowing to idempotent updates
        if old_selection_table is not None:
            check_pxt_table_has_primary_key(old_selection_table, set(["idx_vector"]))

        if not self.allow_existing and old_selection_table is not None:
            raise ValueError(
                f"Table at path {self.table_path} already exists and "
                "allow_existing is set to False."
            )

        if isinstance(select_from, Table):
            selection_table = self._selection_table_from_table(
                select_from=select_from,
                old_selection_table=old_selection_table,
            )
        else:
            selection_table = self._selection_table_from_dataset(
                select_from, old_selection_table
            )

        return selection_table

    def select_in_dataset(
        self, select_from: Dataset | Table, select_count: int, value: str
    ) -> GoldPxtTorchDataset:
        """Select a subset of samples using coresubset selection and return results as a GoldPxtTorchDataset.

        The selection process applies a coresubset selection algorithm on already vectorized
        representations of the data points and stores results in a PixelTable table specified by `table_path`.
        When the chunk attribute is set, the selection is performed in chunks to reduce memory consumption.
        If a reducer is provided, the vectors are reduced in dimension before applying the coresubset selection.

        This is a convenience wrapper that runs `select_in_table` to populate
        (or resume populating) the PixelTable table, then wraps the table into a
        `GoldPxtTorchDataset` for downstream consumption. If `drop_table` is True,
        the table will be removed after the dataset is created.

        Args:
            select_from: Dataset or Table to select from. If a Dataset is provided, each item should be a
                dictionary with at least the `vectorized_key` and `idx` keys after applying the collate_fn.
                If the collate_fn is None, the dataset is expected to directly provide such batches.
                If a Table is provided, it should contain at least the `vectorized_key`, `idx` and `idx_vector` columns.
            select_count: Number of data points to select.
            value: Value to set in the `selection_key` column for selected samples.

        Returns:
            A GoldPxtTorchDataset containing at least the selection information in the `selection_key` key
                and `idx` (index of the sample) and `idx_vector` (index of the vector) keys as well.
        """

        selected_table = self.select_in_table(select_from, select_count, value)

        selected_dataset = GoldPxtTorchDataset(selected_table, keep_cache=True)

        if self.drop_table:
            pxt.drop_table(selected_table)

        return selected_dataset

    def select_in_table(
        self, select_from: Dataset | Table, select_size: int | float, value: str | None
    ) -> Table:
        """Select a subset of samples using coresubset selection and store results in a PixelTable table.

        The selection process applies a coresubset selection algorithm on already vectorized
        representations of the data points and stores results in a PixelTable table specified by `table_path`.
        When the chunk attribute is set, the selection is performed in chunks to reduce memory consumption.
        If a reducer is provided, the vectors are reduced in dimension before applying the coresubset selection.

        This method is idempotent (i.e., failure proof), meaning that if it is called
        multiple times on the same dataset or table, it will restart the selection process
        based on the vectors already present in the PixelTable table.

        Args:
            select_from: Dataset or Table to select from. If a Dataset is provided, each item should be a
                dictionary with at least the `vectorized_key` and `idx` keys after applying the collate_fn.
                If the collate_fn is None, the dataset is expected to directly provide such batches.
                If a Table is provided, it should contain at least the `vectorized_key`, `idx` and `idx_vector` columns.
            select_size: Number or ratio (between 0 and 1) of data points to select.
            value: Value to set in the `selection_key` column for selected samples.

        Returns:
            A PixelTable Table containing at least the selection information in the `selection_key` column
                and `idx` (index of the sample) and `idx_vector` (index of the vector) columns as well.

        Raises:
            ValueError: If select_size is a float not in the range (0.0, 1.0).
        """
        if isinstance(select_size, float):
            if not (0.0 < select_size < 1.0):
                raise ValueError(
                    "When select_size is a float, it must be in the range (0.0, 1.0)."
                )
        else:
            if select_size <= 0:
                raise ValueError("select_size must be a positive integer.")

        selection_table = self.setup_selection_table(select_from)
        if isinstance(select_from, Dataset):
            select_from = selection_table

        assert isinstance(select_from, Table)

        # define the number of element to sample
        total_size = select_from.select(select_from.idx).distinct().count()
        select_count = (
            select_size
            if isinstance(select_size, int)
            else int(select_size * total_size)
        )
        if select_count == 0 and isinstance(select_size, float):
            select_count = 1  # at least one sample

        selection_indices = self.get_selection_indices(
            selection_table, value, self.selection_key
        )
        already_selected_count = len(selection_indices)
        if already_selected_count > 0:
            # if new vectors have been added to the selection table
            # make sure already selected samples are already flagged out
            set_value_to_idx_rows(
                table=selection_table,
                col_expr=get_expr_from_column_name(selection_table, self.selection_key),
                idx_expr=selection_table.idx,
                indices=selection_indices,
                value=value,
            )

        still_to_select_count = select_count - already_selected_count
        if still_to_select_count == 0:
            logger.info(
                f"Selection table already fully filled out for {value} from {self.table_path}"
            )
            return selection_table
        elif still_to_select_count < 0:
            raise ValueError(
                f"Selection table already contains {already_selected_count} selected samples for {value}, "
                f"which is more than the requested {select_count} samples."
            )
        elif self.distribute:
            self._distributed_select(select_from, selection_table, select_count, value)
        else:
            self._sequential_select(select_from, selection_table, select_count, value)

        logger.info(
            f"Selection table populated {
                len(
                    self.get_selection_indices(
                        selection_table, value, self.selection_key
                    )
                )
            } rows with value {value} at {self.table_path}"
        )

        return selection_table

    def _selection_table_from_table(
        self, select_from: Table, old_selection_table: Table | None
    ) -> Table:
        """Create or validate the selection table schema from a PixelTable table.

        This private method sets up the table structure with necessary columns for tracking
        selection status and ensures all rows from the source table are represented.

        Args:
            select_from: The source PixelTable table to select from.
            old_selection_table: Existing selection table if resuming, or None.

        Returns:
            The selection table with proper schema and initial rows.
        """
        minimal_schema = self._MINIMAL_SCHEMA.copy()

        if self.to_keep_schema is not None:
            minimal_schema |= self.to_keep_schema

        if self.label_key is not None:
            minimal_schema[self.label_key] = pxt.String

        selection_table = get_valid_table(
            table=old_selection_table
            if old_selection_table is not None
            else self.table_path,
            minimal_schema=minimal_schema,
            primary_key="idx_vector",
        )

        if self.vectorized_key not in select_from.columns():
            raise ValueError(
                f"Table at path {self.table_path} does not contain "
                f"the required column {self.vectorized_key}."
            )

        selection_table_cols = selection_table.columns()
        if self.selection_key not in selection_table_cols:
            selection_table.add_column(
                if_exists="error", **{self.selection_key: pxt.String}
            )

        if (
            self.include_vectorized_in_table
            and self.vectorized_key not in selection_table_cols
        ):
            idx_vectors = [row["idx_vector"] for row in select_from.sample(1).collect()]
            if not idx_vectors:
                raise ValueError(
                    f"Source table at {self.table_path} is empty, cannot sample vectorized data for schema inference."
                )
            sample = get_sample_row_from_idx(
                select_from,
                idx_key="idx_vector",
                # make sure to take an existing vector
                idx=idx_vectors[0],
                collate_fn=self.collate_fn,
                expected_keys=[self.vectorized_key],
            )

            vectorized_value = sample[self.vectorized_key]
            selection_table.add_column(
                **{
                    self.vectorized_key: pxt.Array[  # type: ignore[misc]
                        vectorized_value.shape, pxt.Float
                    ]
                }
            )

        if selection_table.count() > 0:
            to_select_indices = set(
                [
                    row["idx_vector"]
                    for row in select_from.select(select_from.idx_vector)
                    .distinct()
                    .collect()
                ]
            )
            already_in_selection = set(
                [
                    row["idx_vector"]
                    for row in selection_table.select(selection_table.idx_vector)
                    .distinct()
                    .collect()
                ]
            )
            still_to_select = to_select_indices.difference(already_in_selection)
            if not still_to_select:
                logger.info(
                    f"The selection table is already initialized in {self.table_path}"
                )
                return selection_table

        col_list = ["idx", self.vectorized_key]
        if self.to_keep_schema is not None:
            col_list.extend(list(self.to_keep_schema.keys()))

        if self.selection_key in select_from.columns():
            col_list.append(self.selection_key)

        if self.label_key is not None and self.label_key in select_from.columns():
            col_list.append(self.label_key)

        self._add_rows_to_selection_table_from_dataset(
            select_from=GoldPxtTorchDataset(
                select_from.select(
                    *[
                        get_expr_from_column_name(select_from, col)
                        for col in col_list + ["idx_vector"]
                    ]
                ),
                keep_cache=False,
            ),
            selection_table=selection_table,
            include_vectorized=self.include_vectorized_in_table,
        )

        return selection_table

    def _selection_table_from_dataset(
        self, select_from: Dataset, old_selection_table: Table | None
    ) -> Table:
        """Create or validate the selection table schema from a PyTorch Dataset.

        This private method sets up the table structure with necessary columns
        and validate that the vectorized column is present in the dataset.

        Args:
            select_from: The source PyTorch Dataset to select from.
            old_selection_table: Existing selection table if resuming, or None.

        Returns:
            The selection table with proper schema.
        """
        minimal_schema = self._MINIMAL_SCHEMA.copy()
        if self.to_keep_schema is not None:
            minimal_schema |= self.to_keep_schema

        if self.label_key is not None:
            minimal_schema[self.label_key] = pxt.String

        selection_table = get_valid_table(
            table=old_selection_table
            if old_selection_table is not None
            else self.table_path,
            minimal_schema=minimal_schema,
            primary_key="idx_vector",
        )
        selection_table_cols = selection_table.columns()
        if self.vectorized_key not in selection_table_cols:
            sample = get_dataset_sample_dict(
                select_from,
                collate_fn=self.collate_fn,
                expected=[self.vectorized_key],
            )

            vectorized_value = (
                sample[self.vectorized_key].squeeze(0).detach().cpu().numpy()
            )
            selection_table.add_column(
                **{
                    self.vectorized_key: pxt.Array[  # type: ignore[misc]
                        vectorized_value.shape, pxt.Float
                    ]
                }
            )

        if self.selection_key not in selection_table_cols:
            selection_table.add_column(
                if_exists="error", **{self.selection_key: pxt.String}
            )

        self._add_rows_to_selection_table_from_dataset(
            select_from, selection_table, include_vectorized=True
        )

        return selection_table

    def _add_rows_to_selection_table_from_dataset(
        self,
        select_from: Dataset,
        selection_table: Table,
        include_vectorized: bool = False,
    ) -> None:
        """Add rows from the source dataset to the selection table.

        This private method iterates through the dataset in batches and populates the
        selection table with vectorized data and metadata, skipping already processed samples.

        Args:
            select_from: The source PyTorch Dataset.
            selection_table: The selection table to populate.
            include_vectorized: Whether to include the vectorized data in the selection table.
        """
        dataloader = DataLoader(
            select_from,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            collate_fn=self.collate_fn,
        )

        # the vectors already present in the selection table will be excluded
        # from the batches to process to avoid duplicates and speed up processing
        # as well as the ones with excluded labels if specified
        to_exclude_from_batch = set(
            [
                row["idx_vector"]
                for row in selection_table.select(selection_table.idx_vector).collect()
            ]
        )

        ready_to_insert: list[dict[str, Any]] = []

        for batch_idx, batch in tqdm(
            enumerate(dataloader),
            desc="Initializing rows for the selection table",
            total=(
                None if hasattr(select_from, "__len__") is False else len(dataloader)
            ),
        ):
            # Stop if we've processed enough batches
            if self.max_batches is not None and batch_idx >= self.max_batches:
                break

            if "idx_vector" not in batch:
                starts = batch_idx * self.batch_size
                batch["idx_vector"] = [
                    starts + idx for idx in range(len(batch[self.vectorized_key]))
                ]

            if "idx" not in batch:
                batch["idx"] = batch["idx_vector"]

            if self.exclude_labels is not None and self.label_key in batch:
                assert self.label_key is not None
                to_exclude_from_batch.update(
                    get_indices_with_labels(
                        batch,
                        self.label_key,
                        self.exclude_labels,
                        index_key="idx_vector",
                    )
                )

            # Remove the vectors already present in the selection table to avoid duplicates and speed up processing
            # and as well the ones with excluded labels
            if to_exclude_from_batch:
                batch = filter_batch_from_indices(
                    batch,
                    to_exclude_from_batch,
                    index_key="idx_vector",
                )

                if len(batch) == 0:
                    continue  # all samples already described

            if self.selection_key not in batch:
                batch[self.selection_key] = [
                    None for _ in range(len(batch[self.vectorized_key]))
                ]

            # add newly added vectors to the exclusion set to avoid duplicates in the next batches
            to_exclude_from_batch.update(
                [
                    idx.item() if isinstance(idx, torch.Tensor) else idx
                    for idx in batch["idx_vector"]
                ]
            )
            to_insert_keys = ["idx", self.selection_key]
            if self.to_keep_schema is not None:
                to_insert_keys.extend(list(self.to_keep_schema.keys()))
            if self.label_key is not None and self.label_key in batch:
                to_insert_keys.append(self.label_key)
            if include_vectorized:
                to_insert_keys.append(self.vectorized_key)

            batch_as_list = make_batch_ready_for_table(
                batch,
                to_insert_keys,
                "idx_vector",
            )

            ready_to_insert.extend(batch_as_list)
            if len(ready_to_insert) >= self.min_pxt_insert_size:
                selection_table.insert(ready_to_insert)
                ready_to_insert = []

        if ready_to_insert:
            selection_table.insert(ready_to_insert)

    @staticmethod
    def get_selection_pxt_query(
        table: Table,
        value: str | None,
        selection_key: str,
        label_key: str | None = None,
        label_value: str | None = None,
        idx_key: str = "idx",
    ) -> Query:
        """Get the Pixeltable query to access samples selected with a given value.

        Args:
            table: PixelTable table to query.
            value: Value in the selection_key column to filter selected samples.
            selection_key: Column name used to store selection values.
            label_key: Optional column name used to filter samples by label.
            label_value: Optional label value to filter samples by label.
            idx_key: Column name used to get sample indices.
        """
        idx_col = get_expr_from_column_name(table, idx_key)
        selection_col = get_expr_from_column_name(table, selection_key)
        if label_value is not None and label_key is not None:
            label_col = get_expr_from_column_name(table, label_key)
            query = (selection_col == value) & (label_col == label_value)  # noqa: E712
        else:
            if label_key is not None or label_value is not None:
                raise ValueError("label_key and label_value must be set together.")
            query = selection_col == value  # noqa: E712

        return table.where(query).select(idx_col).distinct()

    @staticmethod
    def get_selection_indices(
        table: Table,
        value: str | None,
        selection_key: str,
        label_key: str | None = None,
        label_value: str | None = None,
        idx_key: str = "idx",
    ) -> set[int]:
        """Get the indices of samples selected with a given value.

        Args:
            table: PixelTable table to query.
            value: Value in the selection_key column to filter selected samples.
            selection_key: Column name used to store selection values.
            label_key: Optional column name used to filter samples by label.
            label_value: Optional label value to filter samples by label.
            idx_key: Column name used to get sample indices.
        """
        return set(
            [
                row[idx_key]
                for row in GoldSelector.get_selection_pxt_query(
                    table,
                    value,
                    selection_key,
                    label_key,
                    label_value,
                    idx_key,
                ).collect()
            ]
        )

    @staticmethod
    def get_selection_count(
        table: Table,
        value: str | None,
        selection_key: str,
        label_key: str | None = None,
        label_value: str | None = None,
        idx_key: str = "idx",
    ) -> int:
        """Get the number of samples selected with a given value.

        Args:
            table: PixelTable table to query.
            value: Value in the selection_key column to filter selected samples.
            selection_key: Column name used to store selection values.
            label_key: Optional column name used to filter samples by label.
            label_value: Optional label value to filter samples by label.
            idx_key: Column name used to get sample indices.
        """
        return GoldSelector.get_selection_pxt_query(
            table,
            value,
            selection_key,
            label_key,
            label_value,
            idx_key,
        ).count()

    def _sequential_select(
        self,
        select_from: Table,
        selection_table: Table,
        select_count: int,
        value: str | None,
    ) -> None:
        """Run sequential (single-process) selection process.

        This private method handles label-stratified selection if a label_key is configured,
        otherwise performs selection on the full dataset. It delegates the actual coresubset
        selection to _label_select.

        Args:
            select_from: The source table with vectorized data.
            selection_table: The table to store selection results.
            select_count: Number of samples to select.
            value: Value to assign to selected samples in the selection_key column.
        """
        if self.label_key is not None:
            label_col = get_expr_from_column_name(selection_table, self.label_key)

            already_selected_count = self.get_selection_count(
                selection_table,
                value,
                self.selection_key,
            )

            # The selection is done iteratively per label
            # The first labels might as well include the next labels in case of multi labels per sample
            # For the first selection loop, the initial target count is limiting the selection
            # (allowing to maximize the idempotence), but if not enough samples have been selected,
            # a new round of selection is done allowing some labels to not be sampled
            # (the selection count ends up to 0 for the label) and the selection
            # is going on ignoring the count selected in the previous round.
            force_all_labels = True
            labels = sorted(
                set(
                    row[self.label_key]
                    for row in selection_table.select(label_col).distinct().collect()
                )
            )
            while already_selected_count < select_count:
                # For the first loop, the initial count is the target and all labels are forced to be selected.
                # For the next loops, only the missing count is targeted
                loop_select_count = (
                    select_count  # force_all_labels means 1st initial selection loop
                    if force_all_labels
                    else select_count
                    - already_selected_count  # otherwise, select only the remaining samples to select after counting the already selected ones for the label
                )
                # During the loop over the labels, the status of each label is computed and this values
                # allows to define how many samples are still to select for the label
                # For the first loop, the initial population is considered as null in order to maximize the idempotence.
                # For the next loops, the already selected samples are the basis for the selection
                label_counts_init = (
                    {label: 0 for label in labels}
                    if force_all_labels
                    else {
                        label: self.get_selection_count(
                            selection_table,
                            value,
                            self.selection_key,
                            self.label_key,
                            label,
                        )
                        for label in labels
                    }
                )

                selection_label_counts = self._compute_label_counts_for_selection(
                    selection_table=selection_table,
                    labels=labels,
                    select_count=loop_select_count,
                    value=value,
                    force_all_labels=force_all_labels,
                )
                for label_idx, (label_value, label_count) in enumerate(
                    selection_label_counts.items()
                ):
                    if label_count == 0:
                        if force_all_labels:
                            raise ValueError(
                                f"Not enough sample for Label '{label_value}' to run the selection"
                            )

                        continue

                    # if the label is already selected for at least the target count defined for the label,
                    # skip the label (over selection is not an issue)
                    label_count_status = (
                        self.get_selection_count(
                            selection_table,
                            value,
                            self.selection_key,
                            self.label_key,
                            label_value,
                        )
                        - label_counts_init[label_value]
                    )
                    if label_count_status >= label_count:
                        continue

                    # select the missing ones based on the current status of the label
                    # in order to reach the target count for the label
                    logger.info(
                        f"Selecting {label_count} samples for label '{label_value}' for value '{value}'"
                    )
                    self._select_label(
                        select_from,
                        selection_table,
                        (
                            label_count
                            if force_all_labels
                            else label_count - label_count_status
                        ),
                        value,
                        label_value=label_value,
                        from_already=force_all_labels,
                    )

                # after the first loop, labels with lower population might are allowed to be sampled again.
                force_all_labels = False
                already_selected_count = self.get_selection_count(
                    selection_table, value=value, selection_key=self.selection_key
                )

        else:
            logger.info(f"Selecting {select_count} samples for value '{value}'")
            self._select_label(
                select_from,
                selection_table,
                select_count,
                value,
            )

    def _select_label(
        self,
        select_from: Table,
        selection_table: Table,
        select_count: int,
        value: str | None,
        label_value: str | None = None,
        from_already: bool = True,
    ) -> None:
        """Perform coresubset selection for a specific label or all data.

        This private method implements chunked coresubset selection.
        It processes data in chunks to manage memory, applies optional dimensionality reduction,
        and updates the selection table with selected samples.

        Args:
            select_from: The source table with vectorized data.
            selection_table: The table to store selection results.
            select_count: Number of samples to select.
            value: Value to assign to selected samples in the selection_key column.
            label_value: Optional label value to filter samples by label.
            from_already: Whether to restart selection from the already selected ones for the label.
                If False, the selection is done to select the full select_count even if some samples have already
                been selected for the label. If True, the selection is done to select only the remaining samples
                 to reach select_count when counting the already selected ones for the label.
        """
        selection_col = get_expr_from_column_name(selection_table, self.selection_key)
        vectorized_col = get_expr_from_column_name(select_from, self.vectorized_key)

        # define the number of samples to select from the already sampled ones for the label
        # when more_than_already is False, the selection is done to select only the remaining samples
        # to reach select_count. When it is True, the selection is done to select the full select_count
        # even if some samples have already been selected
        already_selected_count = self.get_selection_count(
            table=selection_table,
            value=value,
            selection_key=self.selection_key,
            label_key=self.label_key,
            label_value=label_value,
        )

        # validate that there is still enough samples to select from
        current_selected_count = already_selected_count if from_already else 0
        available_samples_for_selection_count = self.get_selection_count(
            table=selection_table,
            value=None,
            selection_key=self.selection_key,
            label_key=self.label_key,
            label_value=label_value,
        )
        if available_samples_for_selection_count < (
            select_count - current_selected_count
        ):
            raise ValueError(
                "Cannot select more unique data points than available in the dataset."
            )

        # The coresubset selection is done from all the vectors (after filtering) of all data points
        # (depending on data, a data point can have multiple vectors).
        # Then, the same data point can be selected multiple times if it has multiple vectors selected.
        # To achieve select_count of unique data points, we loop until we have enough unique data points selected.
        if label_value is not None:
            assert self.label_key is not None
            label_col = get_expr_from_column_name(selection_table, self.label_key)
            available_query = (selection_col == None) & (label_col == label_value)  # noqa: E712 E711
        else:
            available_query = selection_col == None  # noqa: E711
        while current_selected_count < select_count:
            loop_start = (
                current_selected_count  # validate some samples have been selected
            )
            selection_pool = selection_table.where(available_query)

            # when the number of samples to select is the same as the number of samples available for selection,
            # the coresubset selection can be skipped and all the available samples are selected
            still_selectable = (
                selection_pool.select(selection_table.idx).distinct().count()
            )
            if still_selectable == 0:
                if from_already:
                    raise ValueError(
                        "No more sample available for selection, but the selection is not complete. "
                        "This might be due to an issue in the selection process or the data."
                    )

                break

            loop_select_count = select_count - current_selected_count
            if loop_select_count == still_selectable:
                set_value_to_idx_rows(
                    table=selection_table,
                    col_expr=selection_col,
                    idx_expr=selection_table.idx,
                    indices=set(
                        row["idx"]
                        for row in selection_pool.select(selection_table.idx)
                        .distinct()
                        .collect()
                    ),
                    value=value,
                )
                break

            # assign the vectors to chunks for chunked selection
            to_select_vector_indices = [
                row["idx_vector"]
                for row in selection_pool.select(selection_table.idx_vector).collect()
            ]
            chunk_assignment = get_random_chunk_assignment(
                to_chunk=to_select_vector_indices,
                max_chunk_size=self.chunk,
                random_state=self.random_state,
            )
            select_count_per_chunk = split_sampling_among_chunks(
                to_split=loop_select_count,
                chunk_sizes=[len(chunk) for chunk in chunk_assignment],
            )

            # make coresubset selection per chunk
            for chunk_indices, chunk_select_count in zip(
                chunk_assignment, select_count_per_chunk
            ):
                if chunk_select_count == 0:
                    continue

                # update the vector available for selection because the selection from previous chunks
                # might have selected some of the samples in the current chunk, making them no longer
                # available for selection in the current chunk
                chunk_vector_indices_not_selected = [
                    row["idx_vector"]
                    for row in selection_table.where(
                        (selection_table.idx_vector.isin(chunk_indices))
                        & (selection_col == None)  # noqa: E711
                    )
                    .select(selection_table.idx_vector)
                    .collect()
                ]
                if len(chunk_vector_indices_not_selected) == 0:
                    continue

                # load the vectors and the corresponding indices for the chunk
                to_select_from = select_from.where(
                    select_from.idx_vector.isin(chunk_vector_indices_not_selected)
                ).select(select_from.idx, vectorized_col)
                to_select_for_chunk = [
                    (
                        torch.from_numpy(sample[self.vectorized_key]),
                        torch.tensor(sample["idx"]).unsqueeze(0),
                    )
                    for sample in to_select_from.collect()
                ]
                vectors_list, indices_list = map(list, zip(*to_select_for_chunk))
                vectors = torch.stack(vectors_list, dim=0)
                indices = torch.cat(indices_list, dim=0)

                if len(to_select_for_chunk) == chunk_select_count:
                    # take all the indices
                    # It can happen when the selection size is close to the total size
                    coresubset_indices = set(indices.tolist())
                else:
                    # make coresubset selection for the chunk
                    if self.reducer is not None:
                        vectors = (
                            self.reducer.fit_transform(vectors)
                            if isinstance(self.reducer, GoldReductionToolWithFit)
                            else self.reducer.transform(vectors)
                        )

                    coresubset_indices = self._coresubset_selection(
                        vectors, chunk_select_count, indices
                    )

                # update table with selected indices
                set_value_to_idx_rows(
                    table=selection_table,
                    col_expr=selection_col,
                    idx_expr=selection_table.idx,
                    indices=coresubset_indices,
                    value=value,
                )

                current_selected_count = self.get_selection_count(
                    table=selection_table,
                    value=value,
                    selection_key=self.selection_key,
                    label_key=self.label_key,
                    label_value=label_value,
                )
                if not from_already:
                    current_selected_count -= already_selected_count  # remove the initial selection count to follow the current one

            if current_selected_count == loop_start:
                raise ValueError(
                    "No new samples were selected in the iteration, but the selection is not complete. "
                    "This might be due to an issue in the selection process or the data."
                )

    def _distributed_select(
        self,
        select_from: Table,
        selection_table: Table,
        select_count: int,
        value: str | None,
    ) -> None:
        """Run distributed selection process (not implemented).

        Args:
            select_from: The source table with vectorized data.
            selection_table: The table to store selection results.
            select_count: Number of samples to select.
            value: Value to assign to selected samples in the selection_key column.

        Raises:
            NotImplementedError: Always raised as distributed mode is not yet implemented.
        """
        raise NotImplementedError("Distributed selection is not implemented yet.")

    def _coresubset_selection(
        self, x: torch.Tensor, select_count: int, indices: torch.Tensor
    ) -> set[int]:
        """Apply the selection algorithm.

        Args:
            x: Input vectors to select from.
            select_count: Number of vectors to select.
            indices: Original indices corresponding to each vector.

        Returns:
            Set of selected indices from the original index space.
        """
        selected_indices = self.selection_tool.select(x, select_count)

        return set(indices[torch.tensor(selected_indices)].tolist())

    def _compute_label_counts_for_selection(
        self,
        selection_table: Table,
        labels: list[str],
        select_count: int,
        value: str | None = None,
        force_all_labels: bool = True,
    ) -> dict[str, int]:
        assert self.label_key is not None
        # When all labels are required, all the not selected samples and the selected samples for the current
        # value are included in the count of the label.
        # This allows to maximize the idempotence of the selection process
        label_counts = {}
        for label in labels:
            count = self.get_selection_count(
                table=selection_table,
                value=None,
                selection_key=self.selection_key,
                label_key=self.label_key,
                label_value=label,
            )
            if force_all_labels:
                count += self.get_selection_count(
                    table=selection_table,
                    value=value,
                    selection_key=self.selection_key,
                    label_key=self.label_key,
                    label_value=label,
                )

            if count > 0:
                label_counts[label] = count
            elif force_all_labels:
                raise ValueError(
                    f"Label '{label}' has no more selectable sample for value '{value}', but force_all_labels is True. "
                    f"This might be due to an issue in the selection process or the data."
                )

        label_ratios = {
            value: ratio
            for value, ratio in zip(
                label_counts.keys(),
                get_ratios_for_counts(list(label_counts.values())),
            )
        }

        return get_sampling_count_from_ratios(
            label_ratios, select_count, force_non_zero=force_all_labels
        )



from typing import Any

import torch
from torchvision.transforms.v2 import Transform


class PatchifyImageMask(Transform):
    """Patchify a binary mask in order to align it with the tokens of a ViT model.

    Attributes:
        patch_size: The size of the patches to create.
        match_ratio: The ratio of non zero pixels in a patch to consider it as a valid token for the class.
    """

    def __init__(self, patch_size: int = 16, match_ratio: float = 0.5) -> None:
        super().__init__()
        if patch_size <= 0:
            raise ValueError("patch_size must be a positive integer")
        self.patch_size = patch_size

        if not 0.0 <= match_ratio <= 1.0:
            raise ValueError("match_ratio must be between 0 and 1")

        self.match_ratio = match_ratio

    def transform(self, inpt: Any, params: dict[str, Any] | None = None) -> Any:
        """Patchify a binary mask in order to align it with the tokens of a ViT model.

        Args:
            inpt: A binary mask tensor of shape (B, 1, H, W) where B is the batch size, H and W are the height and width of the image.
            params: A dictionary of parameters (not used in this transform).

        Returns:
            A tensor of shape (B, N) where N is the number of patches (H // patch_size) * (W // patch_size)
            and each value is 1 if the ratio of non zero pixels in the patch is greater than match_ratio, otherwise 0.
        """
        if inpt.ndim != 4:
            raise ValueError("Input tensor must have shape (B, C, H, W)")

        B, C, H, W = inpt.shape
        if C != 1:
            raise ValueError("Input mask must have a single channel (C=1)")

        if H % self.patch_size != 0 or W % self.patch_size != 0:
            raise ValueError("H and W must be divisible by patch_size")

        if not torch.all((inpt == 0) | (inpt == 1)):
            raise ValueError("Input must be a binary mask with values 0 or 1")

        pixels_in_patch = self.patch_size * self.patch_size

        # Patchify over spatial dimensions (H, W) while keeping batch and channel dimensions
        patches = inpt.unfold(2, self.patch_size, self.patch_size).unfold(
            3, self.patch_size, self.patch_size
        )
        # patches shape: (B, 1, H//ps, W//ps, ps, ps)
        patches = patches.contiguous().view(B, -1, pixels_in_patch)
        # compute the ratio of non zero pixels in each patch
        ratios = patches.sum(dim=-1) / pixels_in_patch

        return (ratios > self.match_ratio).float().unsqueeze(1)



from abc import abstractmethod, ABC
from typing_extensions import assert_never
from typing import Dict, List, Callable, Any

from dataclasses import dataclass
from enum import Enum

import torch


class GoldEmbeddingTool(ABC):
    """Abstract base class for embedding extraction from models.

    This class defines the interface for embedding tools that can embed and optionally
    fuse embeddings from models. Implementations should provide specific mechanisms for
    extracting embeddings from different types of models (e.g., PyTorch, multimodal).
    """

    @abstractmethod
    def embed(self, *args: Any, **kwargs: Any) -> dict[str, torch.Tensor]:
        """Embed the input data using the model.

        Returns: Dictionary mapping layer names to their embedding tensors.
        """

    @abstractmethod
    def embed_and_fuse(self, *args: Any, **kwargs: Any) -> torch.Tensor:
        """Embed and fuse the input data using the model.

        Returns: Fused embedding tensor.
        """


class EmbeddingFusionStrategy(Enum):
    """Strategies to fuse embeddings from multiple layers.

    CONCAT: Concatenate embeddings along the channel dimension.
    ADD: Element-wise addition of embeddings.
    AVERAGE: Element-wise average of embeddings.
    MAX: Element-wise maximum of embeddings.
    """

    CONCAT = "concat"
    ADD = "add"
    AVERAGE = "average"
    MAX = "max"


@dataclass
class TorchGoldEmbeddingToolConfig:
    """Configuration for the TorchGoldEmbeddingTool.

    Attributes:
        model: The PyTorch model from which to extract embeddings.
        layers: List of layer names or a dictionary mapping group names to lists of layer names.
            If None, the last layer of the model is used.
        layer_fusion: Strategy to fuse embeddings from multiple layers within the same group.
        group_fusion: Strategy to fuse embeddings from different groups.
    """

    model: torch.nn.Module
    layers: list[str] | dict[str, list[str]] | None = None
    layer_fusion: EmbeddingFusionStrategy = EmbeddingFusionStrategy.CONCAT
    group_fusion: EmbeddingFusionStrategy = EmbeddingFusionStrategy.CONCAT


class GoldEmbeddingFusionTool:
    """Embedding fusion from multiple layers and groups.

    Attributes:
        layer_fusion: Strategy to fuse embeddings from multiple layers within the same group.
        group_fusion: Strategy to fuse embeddings from different groups.
    """

    def __init__(
        self,
        layer_fusion: EmbeddingFusionStrategy = EmbeddingFusionStrategy.CONCAT,
        group_fusion: EmbeddingFusionStrategy = EmbeddingFusionStrategy.CONCAT,
    ) -> None:
        """Initialize the GoldEmbeddingFusionTool.

        Args:
            layer_fusion: Strategy to fuse embeddings from multiple layers within the same group.
                Defaults to CONCAT.
            group_fusion: Strategy to fuse embeddings from different groups. Defaults to CONCAT.
        """
        self.layer_fusion = layer_fusion
        self.group_fusion = group_fusion

    @staticmethod
    def fuse_tensors(
        tensors: List[torch.Tensor],
        strategy: EmbeddingFusionStrategy,
    ) -> torch.Tensor:
        """Fuse a list of tensors.

        The tensors are at least expected to have the shape (B, C), until (B, C, D, H, W) if sizes after the
        channel dimension differ. In this case, all tensors are interpolated to the largest size.

        Args:
            tensors: List of tensors to be fused.
            strategy: Strategy to fuse the tensors.

        Returns: Fused tensors.

        Raises:
            ValueError: If the tensors have a different number of dimensions.
        """
        ndims = set(f.ndim for f in tensors)
        if len(ndims) != 1:
            raise ValueError("All embeddings must have the same number of dimensions.")

        ndim = list(ndims)[0]
        if ndim > 2:
            max_size = tuple(
                max(sizes) for sizes in zip(*(f.shape[2:] for f in tensors))
            )
            # Interpolate all tensors to the largest size
            mode = "linear" if ndim == 3 else ("bilinear" if ndim == 4 else "trilinear")
            tensors = [
                (
                    torch.nn.functional.interpolate(
                        embedding,
                        size=max_size,
                        mode=mode,
                    )
                    if embedding.shape[2:] != max_size
                    else embedding
                )
                for embedding in tensors
            ]

        if strategy is EmbeddingFusionStrategy.CONCAT:
            return torch.cat(tensors, dim=1)
        elif strategy is EmbeddingFusionStrategy.ADD:
            return torch.stack(tensors, dim=0).sum(dim=0)
        elif strategy is EmbeddingFusionStrategy.AVERAGE:
            return torch.stack(tensors, dim=0).mean(dim=0)
        elif strategy is EmbeddingFusionStrategy.MAX:
            return torch.stack(tensors, dim=0).max(dim=0).values
        else:
            assert_never(strategy)

    def fuse_embeddings(
        self,
        x: dict[str, torch.Tensor],
        layers: list[str] | dict[str, list[str]],
    ) -> torch.Tensor:
        """Fuse embeddings from multiple layers and groups.

        Args:
            x: Dictionary mapping layer names to embedding tensors.
            layers: List of layer names or a dictionary mapping group names to lists of layer names.

        Returns: Fused embedding tensor.
        """
        # list of layers are fused by layer_fusion strategy
        if isinstance(layers, list):
            return self.fuse_tensors([x[name] for name in layers], self.layer_fusion)

        # groups of layers are fused by layer_fusion strategy,
        # then all groups are fused by group_fusion strategy
        fused_groups = []
        for group, layer_names in layers.items():
            if len(layer_names) == 1:
                fused_groups.append(x[layer_names[0]])
            else:
                fused_groups.append(
                    self.fuse_tensors(
                        [x[name] for name in layer_names], self.layer_fusion
                    )
                )
        return self.fuse_tensors(fused_groups, self.group_fusion)


class TorchGoldEmbeddingTool(GoldEmbeddingTool):
    """Embedding tool for PyTorch models.

    Once initialized, the tool registers forward hooks on the specified layers of the model.
    When the model processes input data, the hooks capture the outputs of these layers.
    The extracted embeddings can then be fused according to the specified strategies.

    The model and layers cannot be changed after initialization. The embedding fusion can be changed.

    Attributes:
        _model: The PyTorch model from which to extract embeddings.
        fusion: GoldEmbeddingFusionTool instance to handle embedding fusion.
        _layers: List of layer names or a dictionary mapping group names to lists of layer names.
        _hooks: Dictionary mapping layer names to their corresponding forward hook handles.
        _embeddings: Dictionary to store extracted embeddings.
    """

    def __init__(
        self,
        config: TorchGoldEmbeddingToolConfig,
    ) -> None:
        """Initialize the TorchGoldEmbeddingTool.

        Args:
            config: Configuration object containing the model, layers, and fusion strategies.
        """
        self._model = config.model
        self.fusion = GoldEmbeddingFusionTool(
            layer_fusion=config.layer_fusion,
            group_fusion=config.group_fusion,
        )

        self._layers: list[str] | dict[str, list[str]]
        self._hooks: dict[str, torch.utils.hooks.RemovableHandle]
        self._embeddings: dict[str, torch.Tensor]

        self._register_layers(config.layers)

    @property
    def model(self) -> torch.nn.Module:
        """The PyTorch model from which to extract embeddings."""
        return self._model

    @property
    def layers(self) -> list[str] | dict[str, list[str]]:
        """The layers from which to extract embeddings.

        It also indicates the grouping of layers for fusion.
        """
        return self._layers

    def embed(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        """Embed the input data using the model.

        Args:
            x: Input data tensor to be processed by the model.

        Returns: Dictionary mapping layer names to their embedding tensors.
        """
        self._embeddings = {}
        self._model(x)
        return self._embeddings

    def embed_and_fuse(self, x: torch.Tensor) -> torch.Tensor:
        """Embed and fuse the input data using the model.

        Args:
            x: Input data tensor to be processed by the model.

        Returns: Fused embedding tensor.
        """
        embeddings = self.embed(x)
        return self.fusion.fuse_embeddings(embeddings, self._layers)

    def __del__(self):
        """Remove all registered hooks when the embedder is deleted."""
        for handle in self._hooks.values():
            handle.remove()

    def _register_layers(self, layers: List[str] | Dict[str, List[str]] | None) -> None:
        """Register forward hooks on the specified layers of the model.

        Args:
            layers: List of layer names or a dictionary mapping group names to lists of layer names.
                If None, the last layer of the model is used.
        """

        named_modules = list(self._model.named_modules())
        if layers is None:
            layer_names = [named_modules[-1][0]]  # last layer
            layers = layer_names
        elif isinstance(layers, dict):
            layer_names = [name for names in layers.values() for name in names]
        elif isinstance(layers, list):
            layer_names = layers
        else:
            assert_never(layers)

        self._layers = layers

        self._embeddings = {}

        def _get_hook(
            name: str,
        ) -> Callable[[torch.nn.Module, torch.Tensor, torch.Tensor], None]:
            def hook(
                module: torch.nn.Module,
                input: torch.Tensor,
                output: torch.Tensor,
            ) -> None:
                self._embeddings[name] = output.detach()

            return hook

        self._hooks = {
            name: module.register_forward_hook(_get_hook(name))
            for name, module in named_modules
            if name in layer_names
        }

        not_found = set(layer_names).difference(set(self._hooks.keys()))
        if not_found:
            raise ValueError(f"Layers not found in the model: {not_found}")


class MultiModalTorchGoldEmbeddingTool(GoldEmbeddingTool):
    """Embedding tool for multimodal data using PyTorch.

    Each modality has its own TorchGoldEmbeddingTool defined by its own configuration.
    This allows for processing different types of input data (e.g., images, text, audio)
    with different models and then fusing their embeddings.

    Attributes:
        embedders: Dictionary mapping modality names to their TorchGoldEmbeddingTool instances.
        strategy: Strategy for fusing embeddings from different modalities.
    """

    def __init__(
        self,
        configs: Dict[str, TorchGoldEmbeddingToolConfig],
        strategy: EmbeddingFusionStrategy = EmbeddingFusionStrategy.CONCAT,
    ) -> None:
        """Initialize the multimodal embedding tool.

        Args:
            configs: Dictionary mapping modality names to their TorchGoldEmbeddingToolConfig.
            strategy: Strategy to use for fusing embeddings from different modalities. Defaults to CONCAT.
        """
        self.embedders = {
            modality: TorchGoldEmbeddingTool(config)
            for modality, config in configs.items()
        }
        self.strategy = strategy

    def embed_and_fuse(self, x: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Embed and fuse multimodal input data.

        Args:
            x: Dictionary mapping modality names to their input tensors.

        Returns:
            Fused embedding tensor combining all modalities.
        """
        return GoldEmbeddingFusionTool.fuse_tensors(
            [
                tool.embed_and_fuse(x[modality])
                for modality, tool in self.embedders.items()
            ],
            self.strategy,
        )

    def embed(self, x: Dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        """Embed multimodal input data without fusing.

        Args:
            x: Dictionary mapping modality names to their input tensors.

        Returns:
            Dictionary mapping "{modality}.{layer}" to their embedding tensors.
        """
        per_modality = {
            modality: tool.embed(x[modality])
            for modality, tool in self.embedders.items()
        }

        return {
            f"{modality}.{layer}": embedding
            for modality, embeddings in per_modality.items()
            for layer, embedding in embeddings.items()
        }
